from __future__ import annotations

import json
from inspect import isclass
from typing import Any

import pydantic
from packaging import version
from pydantic import BaseModel
from typing_extensions import TypeGuard

try:
    from pydantic.v1 import BaseModel as V1BaseModel
except ImportError:
    V1BaseModel = None

is_pydantic_v1 = version.parse(pydantic.__version__).major == 1


def _is_pydantic_v1_basemodel(type_: type) -> TypeGuard[type[BaseModel]]:
    return V1BaseModel is not None and issubclass(type_, V1BaseModel)


def _is_pydantic_v1_basemodel_instance(v: object) -> TypeGuard[BaseModel]:
    return V1BaseModel is not None and isinstance(v, V1BaseModel)


def is_pydantic_basemodel(type_: object) -> TypeGuard[type[BaseModel]]:
    """Check if a type is a Pydantic BaseModel."""
    return isclass(type_) and (issubclass(type_, BaseModel) or _is_pydantic_v1_basemodel(type_))


def is_pydantic_basemodel_instance(v: object) -> TypeGuard[BaseModel]:
    return isinstance(v, BaseModel) or _is_pydantic_v1_basemodel_instance(v)


def get_pydantic_output_structure(model: type[BaseModel]) -> str:
    if is_pydantic_v1 or _is_pydantic_v1_basemodel(model):
        return model.schema_json()
    else:
        return json.dumps(model.model_json_schema())  # pyright: ignore[reportAttributeAccessIssue]


def parse_pydantic_model(model: type[BaseModel], json_str: str) -> BaseModel:
    if is_pydantic_v1 or _is_pydantic_v1_basemodel(model):
        return model.parse_raw(json_str)
    else:
        return model.model_validate_json(json_str)  # pyright: ignore[reportAttributeAccessIssue]


def construct_pydantic_model(model: type[BaseModel], /, **kwargs: Any) -> BaseModel:
    if is_pydantic_v1 or _is_pydantic_v1_basemodel(model):
        return model.construct(**kwargs)
    else:
        return model.model_construct(**kwargs)  # pyright: ignore[reportAttributeAccessIssue]


def get_pydantic_model_dict(model: BaseModel) -> dict[str, Any]:
    if is_pydantic_v1 or _is_pydantic_v1_basemodel_instance(model):
        return model.dict()
    else:
        return model.model_dump()  # pyright: ignore[reportAttributeAccessIssue]


def get_pydantic_model_json(model: BaseModel) -> str:
    if is_pydantic_v1 or _is_pydantic_v1_basemodel_instance(model):
        return model.json()
    else:
        return model.model_dump_json()  # pyright: ignore[reportAttributeAccessIssue]



"""
@generated by mypy-protobuf.  Do not edit manually!
isort:skip_file
"""

from abc import (
    ABCMeta,
    abstractmethod,
)
from chalk._gen.chalk.server.v1.manager_pb2 import (
    GetClusterEnvironmentsRequest,
    GetClusterEnvironmentsResponse,
)
from grpc import (
    Channel,
    Server,
    ServicerContext,
    UnaryUnaryMultiCallable,
)

class ManagerServiceStub:
    def __init__(self, channel: Channel) -> None: ...
    GetClusterEnvironments: UnaryUnaryMultiCallable[
        GetClusterEnvironmentsRequest,
        GetClusterEnvironmentsResponse,
    ]
    """If any checks fail, this request fails."""

class ManagerServiceServicer(metaclass=ABCMeta):
    @abstractmethod
    def GetClusterEnvironments(
        self,
        request: GetClusterEnvironmentsRequest,
        context: ServicerContext,
    ) -> GetClusterEnvironmentsResponse:
        """If any checks fail, this request fails."""

def add_ManagerServiceServicer_to_server(servicer: ManagerServiceServicer, server: Server) -> None: ...



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: chalk/artifacts/v1/cdc.proto
# Protobuf Python Version: 4.25.3
"""Generated protocol buffer code."""

from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(
    b'\n\x1c\x63halk/artifacts/v1/cdc.proto\x12\x12\x63halk.artifacts.v1"?\n\x11\x43\x44\x43TableReference\x12\x12\n\x04name\x18\x01 \x01(\tR\x04name\x12\x16\n\x06schema\x18\x02 \x01(\tR\x06schema"u\n\tCDCSource\x12)\n\x10integration_name\x18\x01 \x01(\tR\x0fintegrationName\x12=\n\x06tables\x18\x02 \x03(\x0b\x32%.chalk.artifacts.v1.CDCTableReferenceR\x06tablesB\x8c\x01\n\x16\x63om.chalk.artifacts.v1B\x08\x43\x64\x63ProtoP\x01\xa2\x02\x03\x43\x41X\xaa\x02\x12\x43halk.Artifacts.V1\xca\x02\x12\x43halk\\Artifacts\\V1\xe2\x02\x1e\x43halk\\Artifacts\\V1\\GPBMetadata\xea\x02\x14\x43halk::Artifacts::V1b\x06proto3'
)

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, "chalk.artifacts.v1.cdc_pb2", _globals)
if _descriptor._USE_C_DESCRIPTORS == False:
    _globals["DESCRIPTOR"]._options = None
    _globals[
        "DESCRIPTOR"
    ]._serialized_options = b"\n\026com.chalk.artifacts.v1B\010CdcProtoP\001\242\002\003CAX\252\002\022Chalk.Artifacts.V1\312\002\022Chalk\\Artifacts\\V1\342\002\036Chalk\\Artifacts\\V1\\GPBMetadata\352\002\024Chalk::Artifacts::V1"
    _globals["_CDCTABLEREFERENCE"]._serialized_start = 52
    _globals["_CDCTABLEREFERENCE"]._serialized_end = 115
    _globals["_CDCSOURCE"]._serialized_start = 117
    _globals["_CDCSOURCE"]._serialized_end = 234
# @@protoc_insertion_point(module_scope)



import collections
import collections.abc
import dataclasses
import warnings
from typing import Any, Dict, List, Mapping, Optional, Sequence, Tuple, Union, cast

import pyarrow as pa

from chalk.features import Features, TPrimitive, ensure_feature
from chalk.features._encoding.json import FeatureEncodingOptions, unstructure_primitive_to_json
from chalk.features.feature_field import FeatureNotFoundException
from chalk.utils.json import TJSON


@dataclasses.dataclass
class InputEncodeOptions:
    encode_structs_as_objects: bool
    """c.f. encode_structs_as_objects in chalk.features._encoding.json.FeatureEncodingOptions"""

    json_encode: bool
    """
    If True, encode feature values into JSON. Specifically, structs/datetimes/bytes/etc. are encoded
    into a JSON-compatible format.
    If False, encode features values into a 'primitive' type (TPrimitive) but don't apply json encoding.
    This is needed for the HTTP client which sends feature data in JSON requests. However, the GRPC
    client transmits Arrow data which supports a richer set of types.
    """


HTTP_ENCODE_OPTIONS = InputEncodeOptions(json_encode=True, encode_structs_as_objects=False)
GRPC_ENCODE_OPTIONS = InputEncodeOptions(json_encode=False, encode_structs_as_objects=True)


def _recursive_unstructure_primitive_to_json(val: TPrimitive) -> TJSON:
    if isinstance(val, dict):
        return {cast(str, k): _recursive_unstructure_primitive_to_json(v) for k, v in val.items()}
    elif isinstance(val, (list, tuple)):
        return [_recursive_unstructure_primitive_to_json(x) for x in val]
    else:
        return unstructure_primitive_to_json(val)


def validate_iterable_values_in_mapping(inputs: Mapping[str, Sequence[Any]], method_name: Optional[str] = None):
    """
    If a method expects inputs of the form Mapping[str, Sequence[Any]], this function will confirm that the values are in fact sequences.
    In particular, because strings are considered sequences, an input `{"user.name": "Raphael"}` will typecheck but then be converted into a list of
    seven users "R", "a", ...
    :param inputs:
    :param method_name:
    :return:
    """
    try:
        import polars as pl
    except ImportError:
        pl = None

    if not isinstance(inputs, collections.abc.Mapping):  # pyright: ignore[reportUnnecessaryIsInstance]
        # Skip this logic for DataFrames, tables, etc.
        return
    for k, vv in inputs.items():
        if isinstance(vv, pa.Array):
            continue
        if pl is not None and isinstance(vv, pl.Series):
            continue

        function_text = "This function"
        if method_name is not None:
            function_text = f"The function {method_name}"

        if not isinstance(vv, collections.abc.Sequence):  # pyright: ignore[reportUnnecessaryIsInstance]
            message = f"""{function_text} accepts a mapping of string keys to Sequence's of values. For key '{k}', got a value of type {type(vv)!r} which is not a Sequence"""
        elif isinstance(vv, str):
            message = f"""{function_text} accepts a mapping of string keys to Sequence's of values. Key '{k}' has a string value which is likely an error. Did you mean to pass in a list of strings instead?"""
        else:
            continue
        warnings.warn(message)
        return


def recursive_encode_bulk_inputs(
    inputs: Mapping[str, Sequence[Any]], options: InputEncodeOptions
) -> Tuple[Dict[str, Union[List[TJSON], pa.Array]], List[str]]:
    all_warnings: List[str] = []
    validate_iterable_values_in_mapping(inputs)
    encoded_inputs: Dict[str, Union[List[TJSON], pa.Array]] = collections.defaultdict(list)
    for wrapped_feature, vv in inputs.items():
        try:
            feature = ensure_feature(wrapped_feature)
        except FeatureNotFoundException:
            fqn = str(wrapped_feature)
            all_warnings.append(
                f"Input '{fqn}' not recognized. Recursively JSON encoding '{fqn}' and requesting anyways"
            )
            if options.json_encode:
                encoded_inputs[str(fqn)] = [_recursive_unstructure_primitive_to_json(v) for v in vv]
            else:
                encoded_inputs[str(fqn)] = list(iter(vv))
            continue

        if feature.is_has_many:
            for v in vv:
                if not isinstance(v, list):
                    raise TypeError(f"has-many feature '{feature.fqn}' must be a list, but got {type(v).__name__}")

                has_many_result: List[Dict[str, TJSON]] = []
                assert feature.joined_class is not None
                foreign_namespace = feature.joined_class.namespace
                for item in v:
                    # The value can be either a feature instance or a dict
                    if isinstance(item, Features):
                        item = dict(item.items())
                    if not isinstance(item, dict):
                        raise TypeError(
                            (
                                f"Has-many feature '{feature.root_fqn}' must be a list of dictionaries or feature set instances, "
                                f"but got a list of `{type(item).__name__}`"
                            )
                        )
                    # Prepend the namespace onto the dict keys, if it's not already there
                    item = {
                        k if str(k).startswith(f"{foreign_namespace}.") else f"{foreign_namespace}.{str(k)}": sub_v
                        for (k, sub_v) in item.items()
                    }
                    result, inner_warnings = recursive_encode_inputs(item, options=options)
                    all_warnings.extend(inner_warnings)
                    has_many_result.append(result)

                encoded_inputs[feature.root_fqn].append(has_many_result)
        elif feature.is_has_one:
            assert feature.joined_class is not None
            foreign_namespace = feature.joined_class.namespace
            for v in vv:
                # The value can be either a feature instance or a dict
                if isinstance(v, Features):
                    v = dict(v.items())
                if not isinstance(v, dict):
                    raise TypeError(
                        (
                            f"Has-one feature '{feature.root_fqn}' must be a list of dictionaries or feature set instances, "
                            f"but got a list of `{type(v).__name__}`"
                        )
                    )
                # Prepend the namespace onto the dict keys, if needed
                v = {
                    k if str(k).startswith(f"{foreign_namespace}.") else f"{foreign_namespace}.{str(k)}": sub_v
                    for (k, sub_v) in v.items()
                }
                has_one_values, inner_warnings = recursive_encode_inputs(v, options=options)
                all_warnings.extend(inner_warnings)
                # Flatten the has-one inputs onto the encoded inputs dict -- similar to how input
                # features
                root_parts = feature.root_fqn.split(".")
                for k, encoded_v in has_one_values.items():
                    # Chop off the namespace from the nested features, as the
                    # namespace is implied by the has-one feature
                    root_fqn = ".".join((*root_parts, *k.split(".")[1:]))
                    encoded_inputs[root_fqn].append(encoded_v)
        elif isinstance(vv, pa.Array):
            if options.json_encode:
                assert not isinstance(vv, (int, str, Sequence))
                raise ValueError(
                    f"The feature '{wrapped_feature}' contains an invalid value. Pyarrow arrays are only supported by the GRPC Chalk client. Cannot send a pyarrow array containing elements {vv.type} over the HTTP Chalk Client."
                )
            encoded_inputs[feature.root_fqn] = vv
        else:
            for v in vv:
                if feature.primary:
                    if not isinstance(v, (int, str)):
                        raise TypeError(
                            f"Input '{v}' for primary feature {feature.root_fqn} must be of type int or str"
                        )
                if options.json_encode:
                    if isinstance(v, pa.Scalar):
                        assert not isinstance(v, (int, str))
                        raise ValueError(
                            f"The feature '{wrapped_feature}' contains an invalid value. Pyarrow arrays are only supported by the GRPC Chalk client. Cannot send a pyarrow array containing elements {v.type} over the HTTP Chalk Client."
                        )
                    converted_value = feature.converter.from_rich_to_json(
                        v,
                        # Allowing missing values because the server could be on a different version of the code that has a default
                        missing_value_strategy="allow",
                        # (pyarrow.RecordBatch.from_pydict expects dict's for struct data, as opposed to an array.)
                        options=FeatureEncodingOptions(encode_structs_as_objects=options.encode_structs_as_objects),
                    )
                else:
                    if isinstance(v, pa.Scalar):
                        converted_value = v
                    else:
                        converted_value = feature.converter.from_rich_to_primitive(
                            v,
                            # Allowing missing values because the server could be on a different version of the code that has a default
                            missing_value_strategy="allow",
                        )

                encoded_inputs[feature.root_fqn].append(converted_value)
    return encoded_inputs, all_warnings


def recursive_encode_inputs(
    inputs: Mapping[str, Any], options: InputEncodeOptions = HTTP_ENCODE_OPTIONS
) -> Tuple[Dict[str, Union[TJSON, pa.Scalar]], List[str]]:
    bulk_result, warnings = recursive_encode_bulk_inputs({k: [v] for k, v in inputs.items()}, options=options)
    return {k: next(iter(v)) for (k, v) in bulk_result.items()}, warnings



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: chalk/common/v1/upload_features.proto
# Protobuf Python Version: 4.25.3
"""Generated protocol buffer code."""

from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from chalk._gen.chalk.common.v1 import chalk_error_pb2 as chalk_dot_common_dot_v1_dot_chalk__error__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(
    b'\n%chalk/common/v1/upload_features.proto\x12\x0f\x63halk.common.v1\x1a!chalk/common/v1/chalk_error.proto":\n\x15UploadFeaturesRequest\x12!\n\x0cinputs_table\x18\x01 \x01(\x0cR\x0binputsTable"p\n\x16UploadFeaturesResponse\x12\x33\n\x06\x65rrors\x18\x01 \x03(\x0b\x32\x1b.chalk.common.v1.ChalkErrorR\x06\x65rrors\x12!\n\x0coperation_id\x18\x02 \x01(\tR\x0boperationIdB\x88\x01\n\x13\x63om.chalk.common.v1B\x13UploadFeaturesProtoP\x01\xa2\x02\x03\x43\x43X\xaa\x02\x0f\x43halk.Common.V1\xca\x02\x0f\x43halk\\Common\\V1\xe2\x02\x1b\x43halk\\Common\\V1\\GPBMetadata\xea\x02\x11\x43halk::Common::V1b\x06proto3'
)

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, "chalk.common.v1.upload_features_pb2", _globals)
if _descriptor._USE_C_DESCRIPTORS == False:
    _globals["DESCRIPTOR"]._options = None
    _globals[
        "DESCRIPTOR"
    ]._serialized_options = b"\n\023com.chalk.common.v1B\023UploadFeaturesProtoP\001\242\002\003CCX\252\002\017Chalk.Common.V1\312\002\017Chalk\\Common\\V1\342\002\033Chalk\\Common\\V1\\GPBMetadata\352\002\021Chalk::Common::V1"
    _globals["_UPLOADFEATURESREQUEST"]._serialized_start = 93
    _globals["_UPLOADFEATURESREQUEST"]._serialized_end = 151
    _globals["_UPLOADFEATURESRESPONSE"]._serialized_start = 153
    _globals["_UPLOADFEATURESRESPONSE"]._serialized_end = 265
# @@protoc_insertion_point(module_scope)



from chalk._gen.chalk.auth.v1 import permissions_pb2 as _permissions_pb2
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from typing import ClassVar as _ClassVar

DESCRIPTOR: _descriptor.FileDescriptor

class GetUnauthedTestEndpointRequest(_message.Message):
    __slots__ = ()
    def __init__(self) -> None: ...

class GetAuthedTestEndpointRequest(_message.Message):
    __slots__ = ()
    def __init__(self) -> None: ...

class GetViewerTestEndpointRequest(_message.Message):
    __slots__ = ()
    def __init__(self) -> None: ...

class GetDataScientistTestEndpointRequest(_message.Message):
    __slots__ = ()
    def __init__(self) -> None: ...

class GetDeveloperTestEndpointRequest(_message.Message):
    __slots__ = ()
    def __init__(self) -> None: ...

class GetAdminTestEndpointRequest(_message.Message):
    __slots__ = ()
    def __init__(self) -> None: ...

class GetOwnerTestEndpointRequest(_message.Message):
    __slots__ = ()
    def __init__(self) -> None: ...

class GetUnauthedTestEndpointResponse(_message.Message):
    __slots__ = ()
    def __init__(self) -> None: ...

class GetAuthedTestEndpointResponse(_message.Message):
    __slots__ = ()
    def __init__(self) -> None: ...

class GetViewerTestEndpointResponse(_message.Message):
    __slots__ = ()
    def __init__(self) -> None: ...

class GetDataScientistTestEndpointResponse(_message.Message):
    __slots__ = ()
    def __init__(self) -> None: ...

class GetDeveloperTestEndpointResponse(_message.Message):
    __slots__ = ()
    def __init__(self) -> None: ...

class GetAdminTestEndpointResponse(_message.Message):
    __slots__ = ()
    def __init__(self) -> None: ...

class GetOwnerTestEndpointResponse(_message.Message):
    __slots__ = ()
    def __init__(self) -> None: ...



from typing import TYPE_CHECKING, Any, Mapping, Optional, Union

from chalk.integrations.named import create_integration_variable, load_integration_variable
from chalk.streams.base import StreamSource
from chalk.utils.duration import Duration

if TYPE_CHECKING:
    from pydantic import BaseModel
else:
    try:
        from pydantic.v1 import BaseModel
    except ImportError:
        from pydantic import BaseModel

_KINESIS_STREAM_NAME_NAME = "KINESIS_STREAM_NAME"
_KINESIS_STREAM_ARN_NAME = "KINESIS_STREAM_ARN"
_KINESIS_REGION_NAME_NAME = "KINESIS_REGION_NAME"
_KINESIS_LATE_ARRIVAL_DEADLINE_NAME = "KINESIS_LATE_ARRIVAL_DEADLINE"
_KINESIS_DEAD_LETTER_QUEUE_STREAM_NAME_NAME = "KINESIS_DEAD_LETTER_QUEUE_STREAM_NAME"
_KINESIS_AWS_ACCESS_KEY_ID_NAME = "KINESIS_AWS_ACCESS_KEY_ID"
_KINESIS_AWS_SECRET_ACCESS_KEY_NAME = "KINESIS_AWS_SECRET_ACCESS_KEY"
_KINESIS_AWS_SESSION_TOKEN_NAME = "KINESIS_AWS_SESSION_TOKEN"
_KINESIS_ENDPOINT_URL_NAME = "KINESIS_ENDPOINT_URL"
_KINESIS_CONSUMER_ROLE_ARN_NAME = "KINESIS_CONSUMER_ROLE_ARN"


class KinesisSource(StreamSource, BaseModel, frozen=True):
    stream_name: Optional[str] = None
    """The name of your stream. Either this or the stream_arn must be specified"""

    stream_arn: Optional[str] = None
    """The ARN of your stream. Either this or the stream_name must be specified"""

    region_name: Optional[str] = None
    """
    AWS region string, e.g. "us-east-2"
    """

    name: Optional[str] = None
    """
    The name of the integration, as configured in your Chalk Dashboard.
    """

    late_arrival_deadline: Duration = "infinity"
    """
    Messages older than this deadline will not be processed.
    """

    dead_letter_queue_stream_name: Optional[str] = None
    """
    Kinesis stream name to send messages when message processing fails
    """

    aws_access_key_id: Optional[str] = None
    """
    AWS access key id credential
    """

    aws_secret_access_key: Optional[str] = None
    """
    AWS secret access key credential
    """

    aws_session_token: Optional[str] = None
    """
    AWS access key id credential
    """

    endpoint_url: Optional[str] = None
    """
    optional endpoint to hit Kinesis server
    """

    consumer_role_arn: Optional[str] = None
    """
    Optional role ARN for the consumer to assume
    """

    def __init__(
        self,
        *,
        region_name: Optional[str] = None,
        stream_name: Optional[str] = None,
        stream_arn: Optional[str] = None,
        late_arrival_deadline: Duration = "infinity",
        dead_letter_queue_stream_name: Optional[str] = None,
        name: Optional[str] = None,
        aws_access_key_id: Optional[str] = None,
        aws_secret_access_key: Optional[str] = None,
        aws_session_token: Optional[str] = None,
        endpoint_url: Optional[str] = None,
        consumer_role_arn: Optional[str] = None,
        integration_variable_override: Optional[Mapping[str, str]] = None,
    ):
        super(KinesisSource, self).__init__(
            name=name,
            stream_name=stream_name
            or load_integration_variable(
                name=_KINESIS_STREAM_NAME_NAME, integration_name=name, override=integration_variable_override
            )
            or KinesisSource.__fields__["stream_name"].default,
            stream_arn=stream_arn
            or load_integration_variable(
                name=_KINESIS_STREAM_ARN_NAME, integration_name=name, override=integration_variable_override
            )
            or KinesisSource.__fields__["stream_arn"].default,
            late_arrival_deadline=late_arrival_deadline
            or load_integration_variable(
                name=_KINESIS_LATE_ARRIVAL_DEADLINE_NAME, integration_name=name, override=integration_variable_override
            )
            or KinesisSource.__fields__["late_arrival_deadline"].default,
            dead_letter_queue_stream_name=dead_letter_queue_stream_name
            or load_integration_variable(
                name=_KINESIS_DEAD_LETTER_QUEUE_STREAM_NAME_NAME,
                integration_name=name,
                override=integration_variable_override,
            )
            or KinesisSource.__fields__["dead_letter_queue_stream_name"].default,
            aws_access_key_id=aws_access_key_id
            or load_integration_variable(
                name=_KINESIS_AWS_ACCESS_KEY_ID_NAME, integration_name=name, override=integration_variable_override
            )
            or KinesisSource.__fields__["aws_access_key_id"].default,
            aws_secret_access_key=aws_secret_access_key
            or load_integration_variable(
                name=_KINESIS_AWS_SECRET_ACCESS_KEY_NAME, integration_name=name, override=integration_variable_override
            )
            or KinesisSource.__fields__["aws_secret_access_key"].default,
            aws_session_token=aws_session_token
            or load_integration_variable(
                name=_KINESIS_AWS_SESSION_TOKEN_NAME, integration_name=name, override=integration_variable_override
            )
            or KinesisSource.__fields__["aws_session_token"].default,
            region_name=region_name
            or load_integration_variable(
                name=_KINESIS_REGION_NAME_NAME, integration_name=name, override=integration_variable_override
            )
            or KinesisSource.__fields__["region_name"].default,
            endpoint_url=endpoint_url
            or load_integration_variable(
                name=_KINESIS_ENDPOINT_URL_NAME, integration_name=name, override=integration_variable_override
            )
            or KinesisSource.__fields__["endpoint_url"].default,
            consumer_role_arn=consumer_role_arn
            or load_integration_variable(
                name=_KINESIS_CONSUMER_ROLE_ARN_NAME, integration_name=name, override=integration_variable_override
            )
            or KinesisSource.__fields__["consumer_role_arn"].default,
        )
        self.registry.append(self)

    def config_to_json(self) -> Any:
        return self.json()

    @property
    def streaming_type(self) -> str:
        return "kinesis"

    @property
    def dlq_name(self) -> Union[str, None]:
        return self.dead_letter_queue_stream_name

    @property
    def stream_or_topic_name(self) -> str:
        assert self.stream_name is not None
        return self.stream_name

    def _recreate_integration_variables(self) -> dict[str, str]:
        return {
            k: v
            for k, v in [
                create_integration_variable(_KINESIS_STREAM_NAME_NAME, self.name, self.stream_name),
                create_integration_variable(_KINESIS_STREAM_ARN_NAME, self.name, self.stream_arn),
                create_integration_variable(_KINESIS_REGION_NAME_NAME, self.name, self.region_name),
                create_integration_variable(_KINESIS_LATE_ARRIVAL_DEADLINE_NAME, self.name, self.late_arrival_deadline),
                create_integration_variable(
                    _KINESIS_DEAD_LETTER_QUEUE_STREAM_NAME_NAME, self.name, self.dead_letter_queue_stream_name
                ),
                create_integration_variable(_KINESIS_AWS_ACCESS_KEY_ID_NAME, self.name, self.aws_access_key_id),
                create_integration_variable(_KINESIS_AWS_SECRET_ACCESS_KEY_NAME, self.name, self.aws_secret_access_key),
                create_integration_variable(_KINESIS_AWS_SESSION_TOKEN_NAME, self.name, self.aws_session_token),
                create_integration_variable(_KINESIS_ENDPOINT_URL_NAME, self.name, self.endpoint_url),
                create_integration_variable(_KINESIS_CONSUMER_ROLE_ARN_NAME, self.name, self.consumer_role_arn),
            ]
            if v is not None
        }



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: chalk/server/v1/log.proto
# Protobuf Python Version: 4.25.3
"""Generated protocol buffer code."""

from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from chalk._gen.chalk.auth.v1 import permissions_pb2 as chalk_dot_auth_dot_v1_dot_permissions__pb2
from google.protobuf import timestamp_pb2 as google_dot_protobuf_dot_timestamp__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(
    b'\n\x19\x63halk/server/v1/log.proto\x12\x0f\x63halk.server.v1\x1a\x1f\x63halk/auth/v1/permissions.proto\x1a\x1fgoogle/protobuf/timestamp.proto"\x9c\x01\n\x08LogEntry\x12\x0e\n\x02id\x18\x01 \x01(\tR\x02id\x12\x1f\n\x08severity\x18\x02 \x01(\tH\x00R\x08severity\x88\x01\x01\x12\x38\n\ttimestamp\x18\x03 \x01(\x0b\x32\x1a.google.protobuf.TimestampR\ttimestamp\x12\x18\n\x07message\x18\x04 \x01(\tR\x07messageB\x0b\n\t_severity"C\n\x19SearchLogEntriesPageToken\x12&\n\x0fnext_page_token\x18\x01 \x01(\tR\rnextPageToken"\x9d\x01\n\x17SearchLogEntriesRequest\x12\x19\n\x05query\x18\x01 \x01(\tH\x00R\x05query\x88\x01\x01\x12N\n\npage_token\x18\x02 \x01(\x0b\x32*.chalk.server.v1.SearchLogEntriesPageTokenH\x01R\tpageToken\x88\x01\x01\x42\x08\n\x06_queryB\r\n\x0b_page_token"\xc3\x01\n\x18SearchLogEntriesResponse\x12:\n\x0blog_entries\x18\x01 \x03(\x0b\x32\x19.chalk.server.v1.LogEntryR\nlogEntries\x12W\n\x0fnext_page_token\x18\x02 \x01(\x0b\x32*.chalk.server.v1.SearchLogEntriesPageTokenH\x00R\rnextPageToken\x88\x01\x01\x42\x12\n\x10_next_page_token2\x83\x01\n\x10LogSearchService\x12o\n\x10SearchLogEntries\x12(.chalk.server.v1.SearchLogEntriesRequest\x1a).chalk.server.v1.SearchLogEntriesResponse"\x06\x90\x02\x01\x80}\x06\x42\x91\x01\n\x13\x63om.chalk.server.v1B\x08LogProtoP\x01Z\x12server/v1;serverv1\xa2\x02\x03\x43SX\xaa\x02\x0f\x43halk.Server.V1\xca\x02\x0f\x43halk\\Server\\V1\xe2\x02\x1b\x43halk\\Server\\V1\\GPBMetadata\xea\x02\x11\x43halk::Server::V1b\x06proto3'
)

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, "chalk.server.v1.log_pb2", _globals)
if _descriptor._USE_C_DESCRIPTORS == False:
    _globals["DESCRIPTOR"]._options = None
    _globals[
        "DESCRIPTOR"
    ]._serialized_options = b"\n\023com.chalk.server.v1B\010LogProtoP\001Z\022server/v1;serverv1\242\002\003CSX\252\002\017Chalk.Server.V1\312\002\017Chalk\\Server\\V1\342\002\033Chalk\\Server\\V1\\GPBMetadata\352\002\021Chalk::Server::V1"
    _globals["_LOGSEARCHSERVICE"].methods_by_name["SearchLogEntries"]._options = None
    _globals["_LOGSEARCHSERVICE"].methods_by_name["SearchLogEntries"]._serialized_options = b"\220\002\001\200}\006"
    _globals["_LOGENTRY"]._serialized_start = 113
    _globals["_LOGENTRY"]._serialized_end = 269
    _globals["_SEARCHLOGENTRIESPAGETOKEN"]._serialized_start = 271
    _globals["_SEARCHLOGENTRIESPAGETOKEN"]._serialized_end = 338
    _globals["_SEARCHLOGENTRIESREQUEST"]._serialized_start = 341
    _globals["_SEARCHLOGENTRIESREQUEST"]._serialized_end = 498
    _globals["_SEARCHLOGENTRIESRESPONSE"]._serialized_start = 501
    _globals["_SEARCHLOGENTRIESRESPONSE"]._serialized_end = 696
    _globals["_LOGSEARCHSERVICE"]._serialized_start = 699
    _globals["_LOGSEARCHSERVICE"]._serialized_end = 830
# @@protoc_insertion_point(module_scope)



from google.protobuf import struct_pb2 as _struct_pb2
from google.protobuf import timestamp_pb2 as _timestamp_pb2
from google.protobuf.internal import containers as _containers
from google.protobuf.internal import enum_type_wrapper as _enum_type_wrapper
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from typing import (
    ClassVar as _ClassVar,
    Iterable as _Iterable,
    Mapping as _Mapping,
    Optional as _Optional,
    Union as _Union,
)

DESCRIPTOR: _descriptor.FileDescriptor

class CloudProviderKind(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
    __slots__ = ()
    CLOUD_PROVIDER_KIND_UNSPECIFIED: _ClassVar[CloudProviderKind]
    CLOUD_PROVIDER_KIND_UNKNOWN: _ClassVar[CloudProviderKind]
    CLOUD_PROVIDER_KIND_GCP: _ClassVar[CloudProviderKind]
    CLOUD_PROVIDER_KIND_AWS: _ClassVar[CloudProviderKind]

CLOUD_PROVIDER_KIND_UNSPECIFIED: CloudProviderKind
CLOUD_PROVIDER_KIND_UNKNOWN: CloudProviderKind
CLOUD_PROVIDER_KIND_GCP: CloudProviderKind
CLOUD_PROVIDER_KIND_AWS: CloudProviderKind

class AWSCloudWatchConfig(_message.Message):
    __slots__ = ("log_group_path", "log_group_paths")
    LOG_GROUP_PATH_FIELD_NUMBER: _ClassVar[int]
    LOG_GROUP_PATHS_FIELD_NUMBER: _ClassVar[int]
    log_group_path: str
    log_group_paths: _containers.RepeatedScalarFieldContainer[str]
    def __init__(
        self, log_group_path: _Optional[str] = ..., log_group_paths: _Optional[_Iterable[str]] = ...
    ) -> None: ...

class AWSSecretManagerConfig(_message.Message):
    __slots__ = ("secret_kms_arn", "secret_tags", "secret_prefix")
    class SecretTagsEntry(_message.Message):
        __slots__ = ("key", "value")
        KEY_FIELD_NUMBER: _ClassVar[int]
        VALUE_FIELD_NUMBER: _ClassVar[int]
        key: str
        value: str
        def __init__(self, key: _Optional[str] = ..., value: _Optional[str] = ...) -> None: ...

    SECRET_KMS_ARN_FIELD_NUMBER: _ClassVar[int]
    SECRET_TAGS_FIELD_NUMBER: _ClassVar[int]
    SECRET_PREFIX_FIELD_NUMBER: _ClassVar[int]
    secret_kms_arn: str
    secret_tags: _containers.ScalarMap[str, str]
    secret_prefix: str
    def __init__(
        self,
        secret_kms_arn: _Optional[str] = ...,
        secret_tags: _Optional[_Mapping[str, str]] = ...,
        secret_prefix: _Optional[str] = ...,
    ) -> None: ...

class GCPSecretReplicationReplica(_message.Message):
    __slots__ = ("location",)
    LOCATION_FIELD_NUMBER: _ClassVar[int]
    location: str
    def __init__(self, location: _Optional[str] = ...) -> None: ...

class GCPRegionConfig(_message.Message):
    __slots__ = ("scope_type",)
    SCOPE_TYPE_FIELD_NUMBER: _ClassVar[int]
    scope_type: str
    def __init__(self, scope_type: _Optional[str] = ...) -> None: ...

class GCPSecretManagerConfig(_message.Message):
    __slots__ = ("secret_region", "replicas")
    SECRET_REGION_FIELD_NUMBER: _ClassVar[int]
    REPLICAS_FIELD_NUMBER: _ClassVar[int]
    secret_region: str
    replicas: _containers.RepeatedCompositeFieldContainer[GCPSecretReplicationReplica]
    def __init__(
        self,
        secret_region: _Optional[str] = ...,
        replicas: _Optional[_Iterable[_Union[GCPSecretReplicationReplica, _Mapping]]] = ...,
    ) -> None: ...

class GCPWorkloadIdentity(_message.Message):
    __slots__ = ("gcp_project_number", "gcp_service_account", "pool_id", "provider_id")
    GCP_PROJECT_NUMBER_FIELD_NUMBER: _ClassVar[int]
    GCP_SERVICE_ACCOUNT_FIELD_NUMBER: _ClassVar[int]
    POOL_ID_FIELD_NUMBER: _ClassVar[int]
    PROVIDER_ID_FIELD_NUMBER: _ClassVar[int]
    gcp_project_number: str
    gcp_service_account: str
    pool_id: str
    provider_id: str
    def __init__(
        self,
        gcp_project_number: _Optional[str] = ...,
        gcp_service_account: _Optional[str] = ...,
        pool_id: _Optional[str] = ...,
        provider_id: _Optional[str] = ...,
    ) -> None: ...

class DockerBuildConfig(_message.Message):
    __slots__ = (
        "builder",
        "push_registry_type",
        "push_registry_tag_prefix",
        "registry_credentials_secret_id",
        "notification_topic",
    )
    BUILDER_FIELD_NUMBER: _ClassVar[int]
    PUSH_REGISTRY_TYPE_FIELD_NUMBER: _ClassVar[int]
    PUSH_REGISTRY_TAG_PREFIX_FIELD_NUMBER: _ClassVar[int]
    REGISTRY_CREDENTIALS_SECRET_ID_FIELD_NUMBER: _ClassVar[int]
    NOTIFICATION_TOPIC_FIELD_NUMBER: _ClassVar[int]
    builder: str
    push_registry_type: str
    push_registry_tag_prefix: str
    registry_credentials_secret_id: str
    notification_topic: str
    def __init__(
        self,
        builder: _Optional[str] = ...,
        push_registry_type: _Optional[str] = ...,
        push_registry_tag_prefix: _Optional[str] = ...,
        registry_credentials_secret_id: _Optional[str] = ...,
        notification_topic: _Optional[str] = ...,
    ) -> None: ...

class ElasticsearchLogConfig(_message.Message):
    __slots__ = ("username", "password", "endpoint")
    USERNAME_FIELD_NUMBER: _ClassVar[int]
    PASSWORD_FIELD_NUMBER: _ClassVar[int]
    ENDPOINT_FIELD_NUMBER: _ClassVar[int]
    username: str
    password: str
    endpoint: str
    def __init__(
        self, username: _Optional[str] = ..., password: _Optional[str] = ..., endpoint: _Optional[str] = ...
    ) -> None: ...

class AWSCloudConfig(_message.Message):
    __slots__ = (
        "account_id",
        "management_role_arn",
        "region",
        "external_id",
        "deprecated_cloud_watch_config",
        "deprecated_secret_manager_config",
        "workload_identity",
        "docker_build_config",
        "elasticsearch_log_config",
        "cloudwatch_config",
        "secretmanager_config",
        "gcp_workload_identity",
    )
    ACCOUNT_ID_FIELD_NUMBER: _ClassVar[int]
    MANAGEMENT_ROLE_ARN_FIELD_NUMBER: _ClassVar[int]
    REGION_FIELD_NUMBER: _ClassVar[int]
    EXTERNAL_ID_FIELD_NUMBER: _ClassVar[int]
    DEPRECATED_CLOUD_WATCH_CONFIG_FIELD_NUMBER: _ClassVar[int]
    DEPRECATED_SECRET_MANAGER_CONFIG_FIELD_NUMBER: _ClassVar[int]
    WORKLOAD_IDENTITY_FIELD_NUMBER: _ClassVar[int]
    DOCKER_BUILD_CONFIG_FIELD_NUMBER: _ClassVar[int]
    ELASTICSEARCH_LOG_CONFIG_FIELD_NUMBER: _ClassVar[int]
    CLOUDWATCH_CONFIG_FIELD_NUMBER: _ClassVar[int]
    SECRETMANAGER_CONFIG_FIELD_NUMBER: _ClassVar[int]
    GCP_WORKLOAD_IDENTITY_FIELD_NUMBER: _ClassVar[int]
    account_id: str
    management_role_arn: str
    region: str
    external_id: str
    deprecated_cloud_watch_config: AWSCloudWatchConfig
    deprecated_secret_manager_config: AWSSecretManagerConfig
    workload_identity: GCPWorkloadIdentity
    docker_build_config: DockerBuildConfig
    elasticsearch_log_config: ElasticsearchLogConfig
    cloudwatch_config: AWSCloudWatchConfig
    secretmanager_config: AWSSecretManagerConfig
    gcp_workload_identity: GCPWorkloadIdentity
    def __init__(
        self,
        account_id: _Optional[str] = ...,
        management_role_arn: _Optional[str] = ...,
        region: _Optional[str] = ...,
        external_id: _Optional[str] = ...,
        deprecated_cloud_watch_config: _Optional[_Union[AWSCloudWatchConfig, _Mapping]] = ...,
        deprecated_secret_manager_config: _Optional[_Union[AWSSecretManagerConfig, _Mapping]] = ...,
        workload_identity: _Optional[_Union[GCPWorkloadIdentity, _Mapping]] = ...,
        docker_build_config: _Optional[_Union[DockerBuildConfig, _Mapping]] = ...,
        elasticsearch_log_config: _Optional[_Union[ElasticsearchLogConfig, _Mapping]] = ...,
        cloudwatch_config: _Optional[_Union[AWSCloudWatchConfig, _Mapping]] = ...,
        secretmanager_config: _Optional[_Union[AWSSecretManagerConfig, _Mapping]] = ...,
        gcp_workload_identity: _Optional[_Union[GCPWorkloadIdentity, _Mapping]] = ...,
    ) -> None: ...

class GCPCloudConfig(_message.Message):
    __slots__ = (
        "project_id",
        "region",
        "management_service_account",
        "docker_build_config",
        "secretmanager_config",
        "region_config",
    )
    PROJECT_ID_FIELD_NUMBER: _ClassVar[int]
    REGION_FIELD_NUMBER: _ClassVar[int]
    MANAGEMENT_SERVICE_ACCOUNT_FIELD_NUMBER: _ClassVar[int]
    DOCKER_BUILD_CONFIG_FIELD_NUMBER: _ClassVar[int]
    SECRETMANAGER_CONFIG_FIELD_NUMBER: _ClassVar[int]
    REGION_CONFIG_FIELD_NUMBER: _ClassVar[int]
    project_id: str
    region: str
    management_service_account: str
    docker_build_config: DockerBuildConfig
    secretmanager_config: GCPSecretManagerConfig
    region_config: GCPRegionConfig
    def __init__(
        self,
        project_id: _Optional[str] = ...,
        region: _Optional[str] = ...,
        management_service_account: _Optional[str] = ...,
        docker_build_config: _Optional[_Union[DockerBuildConfig, _Mapping]] = ...,
        secretmanager_config: _Optional[_Union[GCPSecretManagerConfig, _Mapping]] = ...,
        region_config: _Optional[_Union[GCPRegionConfig, _Mapping]] = ...,
    ) -> None: ...

class CloudConfig(_message.Message):
    __slots__ = ("aws", "gcp")
    AWS_FIELD_NUMBER: _ClassVar[int]
    GCP_FIELD_NUMBER: _ClassVar[int]
    aws: AWSCloudConfig
    gcp: GCPCloudConfig
    def __init__(
        self,
        aws: _Optional[_Union[AWSCloudConfig, _Mapping]] = ...,
        gcp: _Optional[_Union[GCPCloudConfig, _Mapping]] = ...,
    ) -> None: ...

class EnvironmentObjectStorageConfig(_message.Message):
    __slots__ = ("dataset_bucket", "plan_stages_bucket", "source_bundle_bucket")
    DATASET_BUCKET_FIELD_NUMBER: _ClassVar[int]
    PLAN_STAGES_BUCKET_FIELD_NUMBER: _ClassVar[int]
    SOURCE_BUNDLE_BUCKET_FIELD_NUMBER: _ClassVar[int]
    dataset_bucket: str
    plan_stages_bucket: str
    source_bundle_bucket: str
    def __init__(
        self,
        dataset_bucket: _Optional[str] = ...,
        plan_stages_bucket: _Optional[str] = ...,
        source_bundle_bucket: _Optional[str] = ...,
    ) -> None: ...

class Environment(_message.Message):
    __slots__ = (
        "name",
        "project_id",
        "id",
        "team_id",
        "active_deployment_id",
        "worker_url",
        "service_url",
        "branch_url",
        "offline_store_secret",
        "online_store_secret",
        "feature_store_secret",
        "postgres_secret",
        "online_store_kind",
        "emq_uri",
        "vpc_connector_name",
        "kube_cluster_name",
        "branch_kube_cluster_name",
        "engine_kube_cluster_name",
        "shadow_engine_kube_cluster_name",
        "kube_job_namespace",
        "kube_preview_namespace",
        "kube_service_account_name",
        "streaming_query_service_uri",
        "skip_offline_writes_for_online_cached_features",
        "result_bus_topic",
        "online_persistence_mode",
        "metrics_bus_topic",
        "bigtable_instance_name",
        "bigtable_table_name",
        "cloud_account_locator",
        "cloud_region",
        "cloud_tenancy_id",
        "source_bundle_bucket",
        "engine_docker_registry_path",
        "default_planner",
        "additional_env_vars",
        "additional_cron_env_vars",
        "private_pip_repositories",
        "is_sandbox",
        "cloud_provider",
        "cloud_config",
        "spec_config_json",
        "archived_at",
        "metadata_server_metrics_store_secret",
        "query_server_metrics_store_secret",
        "pinned_base_image",
        "cluster_gateway_id",
        "cluster_timescaledb_id",
        "background_persistence_deployment_id",
        "environment_buckets",
        "cluster_timescaledb_secret",
        "grpc_engine_url",
        "kube_cluster_mode",
        "dashboard_url",
    )
    class AdditionalEnvVarsEntry(_message.Message):
        __slots__ = ("key", "value")
        KEY_FIELD_NUMBER: _ClassVar[int]
        VALUE_FIELD_NUMBER: _ClassVar[int]
        key: str
        value: str
        def __init__(self, key: _Optional[str] = ..., value: _Optional[str] = ...) -> None: ...

    class AdditionalCronEnvVarsEntry(_message.Message):
        __slots__ = ("key", "value")
        KEY_FIELD_NUMBER: _ClassVar[int]
        VALUE_FIELD_NUMBER: _ClassVar[int]
        key: str
        value: str
        def __init__(self, key: _Optional[str] = ..., value: _Optional[str] = ...) -> None: ...

    class SpecConfigJsonEntry(_message.Message):
        __slots__ = ("key", "value")
        KEY_FIELD_NUMBER: _ClassVar[int]
        VALUE_FIELD_NUMBER: _ClassVar[int]
        key: str
        value: _struct_pb2.Value
        def __init__(
            self, key: _Optional[str] = ..., value: _Optional[_Union[_struct_pb2.Value, _Mapping]] = ...
        ) -> None: ...

    NAME_FIELD_NUMBER: _ClassVar[int]
    PROJECT_ID_FIELD_NUMBER: _ClassVar[int]
    ID_FIELD_NUMBER: _ClassVar[int]
    TEAM_ID_FIELD_NUMBER: _ClassVar[int]
    ACTIVE_DEPLOYMENT_ID_FIELD_NUMBER: _ClassVar[int]
    WORKER_URL_FIELD_NUMBER: _ClassVar[int]
    SERVICE_URL_FIELD_NUMBER: _ClassVar[int]
    BRANCH_URL_FIELD_NUMBER: _ClassVar[int]
    OFFLINE_STORE_SECRET_FIELD_NUMBER: _ClassVar[int]
    ONLINE_STORE_SECRET_FIELD_NUMBER: _ClassVar[int]
    FEATURE_STORE_SECRET_FIELD_NUMBER: _ClassVar[int]
    POSTGRES_SECRET_FIELD_NUMBER: _ClassVar[int]
    ONLINE_STORE_KIND_FIELD_NUMBER: _ClassVar[int]
    EMQ_URI_FIELD_NUMBER: _ClassVar[int]
    VPC_CONNECTOR_NAME_FIELD_NUMBER: _ClassVar[int]
    KUBE_CLUSTER_NAME_FIELD_NUMBER: _ClassVar[int]
    BRANCH_KUBE_CLUSTER_NAME_FIELD_NUMBER: _ClassVar[int]
    ENGINE_KUBE_CLUSTER_NAME_FIELD_NUMBER: _ClassVar[int]
    SHADOW_ENGINE_KUBE_CLUSTER_NAME_FIELD_NUMBER: _ClassVar[int]
    KUBE_JOB_NAMESPACE_FIELD_NUMBER: _ClassVar[int]
    KUBE_PREVIEW_NAMESPACE_FIELD_NUMBER: _ClassVar[int]
    KUBE_SERVICE_ACCOUNT_NAME_FIELD_NUMBER: _ClassVar[int]
    STREAMING_QUERY_SERVICE_URI_FIELD_NUMBER: _ClassVar[int]
    SKIP_OFFLINE_WRITES_FOR_ONLINE_CACHED_FEATURES_FIELD_NUMBER: _ClassVar[int]
    RESULT_BUS_TOPIC_FIELD_NUMBER: _ClassVar[int]
    ONLINE_PERSISTENCE_MODE_FIELD_NUMBER: _ClassVar[int]
    METRICS_BUS_TOPIC_FIELD_NUMBER: _ClassVar[int]
    BIGTABLE_INSTANCE_NAME_FIELD_NUMBER: _ClassVar[int]
    BIGTABLE_TABLE_NAME_FIELD_NUMBER: _ClassVar[int]
    CLOUD_ACCOUNT_LOCATOR_FIELD_NUMBER: _ClassVar[int]
    CLOUD_REGION_FIELD_NUMBER: _ClassVar[int]
    CLOUD_TENANCY_ID_FIELD_NUMBER: _ClassVar[int]
    SOURCE_BUNDLE_BUCKET_FIELD_NUMBER: _ClassVar[int]
    ENGINE_DOCKER_REGISTRY_PATH_FIELD_NUMBER: _ClassVar[int]
    DEFAULT_PLANNER_FIELD_NUMBER: _ClassVar[int]
    ADDITIONAL_ENV_VARS_FIELD_NUMBER: _ClassVar[int]
    ADDITIONAL_CRON_ENV_VARS_FIELD_NUMBER: _ClassVar[int]
    PRIVATE_PIP_REPOSITORIES_FIELD_NUMBER: _ClassVar[int]
    IS_SANDBOX_FIELD_NUMBER: _ClassVar[int]
    CLOUD_PROVIDER_FIELD_NUMBER: _ClassVar[int]
    CLOUD_CONFIG_FIELD_NUMBER: _ClassVar[int]
    SPEC_CONFIG_JSON_FIELD_NUMBER: _ClassVar[int]
    ARCHIVED_AT_FIELD_NUMBER: _ClassVar[int]
    METADATA_SERVER_METRICS_STORE_SECRET_FIELD_NUMBER: _ClassVar[int]
    QUERY_SERVER_METRICS_STORE_SECRET_FIELD_NUMBER: _ClassVar[int]
    PINNED_BASE_IMAGE_FIELD_NUMBER: _ClassVar[int]
    CLUSTER_GATEWAY_ID_FIELD_NUMBER: _ClassVar[int]
    CLUSTER_TIMESCALEDB_ID_FIELD_NUMBER: _ClassVar[int]
    BACKGROUND_PERSISTENCE_DEPLOYMENT_ID_FIELD_NUMBER: _ClassVar[int]
    ENVIRONMENT_BUCKETS_FIELD_NUMBER: _ClassVar[int]
    CLUSTER_TIMESCALEDB_SECRET_FIELD_NUMBER: _ClassVar[int]
    GRPC_ENGINE_URL_FIELD_NUMBER: _ClassVar[int]
    KUBE_CLUSTER_MODE_FIELD_NUMBER: _ClassVar[int]
    DASHBOARD_URL_FIELD_NUMBER: _ClassVar[int]
    name: str
    project_id: str
    id: str
    team_id: str
    active_deployment_id: str
    worker_url: str
    service_url: str
    branch_url: str
    offline_store_secret: str
    online_store_secret: str
    feature_store_secret: str
    postgres_secret: str
    online_store_kind: str
    emq_uri: str
    vpc_connector_name: str
    kube_cluster_name: str
    branch_kube_cluster_name: str
    engine_kube_cluster_name: str
    shadow_engine_kube_cluster_name: str
    kube_job_namespace: str
    kube_preview_namespace: str
    kube_service_account_name: str
    streaming_query_service_uri: str
    skip_offline_writes_for_online_cached_features: bool
    result_bus_topic: str
    online_persistence_mode: str
    metrics_bus_topic: str
    bigtable_instance_name: str
    bigtable_table_name: str
    cloud_account_locator: str
    cloud_region: str
    cloud_tenancy_id: str
    source_bundle_bucket: str
    engine_docker_registry_path: str
    default_planner: str
    additional_env_vars: _containers.ScalarMap[str, str]
    additional_cron_env_vars: _containers.ScalarMap[str, str]
    private_pip_repositories: str
    is_sandbox: bool
    cloud_provider: CloudProviderKind
    cloud_config: CloudConfig
    spec_config_json: _containers.MessageMap[str, _struct_pb2.Value]
    archived_at: _timestamp_pb2.Timestamp
    metadata_server_metrics_store_secret: str
    query_server_metrics_store_secret: str
    pinned_base_image: str
    cluster_gateway_id: str
    cluster_timescaledb_id: str
    background_persistence_deployment_id: str
    environment_buckets: EnvironmentObjectStorageConfig
    cluster_timescaledb_secret: str
    grpc_engine_url: str
    kube_cluster_mode: str
    dashboard_url: str
    def __init__(
        self,
        name: _Optional[str] = ...,
        project_id: _Optional[str] = ...,
        id: _Optional[str] = ...,
        team_id: _Optional[str] = ...,
        active_deployment_id: _Optional[str] = ...,
        worker_url: _Optional[str] = ...,
        service_url: _Optional[str] = ...,
        branch_url: _Optional[str] = ...,
        offline_store_secret: _Optional[str] = ...,
        online_store_secret: _Optional[str] = ...,
        feature_store_secret: _Optional[str] = ...,
        postgres_secret: _Optional[str] = ...,
        online_store_kind: _Optional[str] = ...,
        emq_uri: _Optional[str] = ...,
        vpc_connector_name: _Optional[str] = ...,
        kube_cluster_name: _Optional[str] = ...,
        branch_kube_cluster_name: _Optional[str] = ...,
        engine_kube_cluster_name: _Optional[str] = ...,
        shadow_engine_kube_cluster_name: _Optional[str] = ...,
        kube_job_namespace: _Optional[str] = ...,
        kube_preview_namespace: _Optional[str] = ...,
        kube_service_account_name: _Optional[str] = ...,
        streaming_query_service_uri: _Optional[str] = ...,
        skip_offline_writes_for_online_cached_features: bool = ...,
        result_bus_topic: _Optional[str] = ...,
        online_persistence_mode: _Optional[str] = ...,
        metrics_bus_topic: _Optional[str] = ...,
        bigtable_instance_name: _Optional[str] = ...,
        bigtable_table_name: _Optional[str] = ...,
        cloud_account_locator: _Optional[str] = ...,
        cloud_region: _Optional[str] = ...,
        cloud_tenancy_id: _Optional[str] = ...,
        source_bundle_bucket: _Optional[str] = ...,
        engine_docker_registry_path: _Optional[str] = ...,
        default_planner: _Optional[str] = ...,
        additional_env_vars: _Optional[_Mapping[str, str]] = ...,
        additional_cron_env_vars: _Optional[_Mapping[str, str]] = ...,
        private_pip_repositories: _Optional[str] = ...,
        is_sandbox: bool = ...,
        cloud_provider: _Optional[_Union[CloudProviderKind, str]] = ...,
        cloud_config: _Optional[_Union[CloudConfig, _Mapping]] = ...,
        spec_config_json: _Optional[_Mapping[str, _struct_pb2.Value]] = ...,
        archived_at: _Optional[_Union[_timestamp_pb2.Timestamp, _Mapping]] = ...,
        metadata_server_metrics_store_secret: _Optional[str] = ...,
        query_server_metrics_store_secret: _Optional[str] = ...,
        pinned_base_image: _Optional[str] = ...,
        cluster_gateway_id: _Optional[str] = ...,
        cluster_timescaledb_id: _Optional[str] = ...,
        background_persistence_deployment_id: _Optional[str] = ...,
        environment_buckets: _Optional[_Union[EnvironmentObjectStorageConfig, _Mapping]] = ...,
        cluster_timescaledb_secret: _Optional[str] = ...,
        grpc_engine_url: _Optional[str] = ...,
        kube_cluster_mode: _Optional[str] = ...,
        dashboard_url: _Optional[str] = ...,
    ) -> None: ...



from __future__ import annotations

import json
import types
from datetime import date, datetime, time, timedelta
from decimal import Decimal
from io import BytesIO
from typing import (
    Any,
    Dict,
    FrozenSet,
    Generic,
    Iterable,
    List,
    Optional,
    Protocol,
    Sequence,
    Set,
    Tuple,
    Type,
    TypeVar,
    Union,
    cast,
    final,
    overload,
)

import dateutil.tz
import numpy as np
import pyarrow as pa
import pyarrow.compute as pc
import pyarrow.feather as pf
from typing_extensions import Final, get_args, get_origin

from chalk._gen.chalk.arrow.v1 import arrow_pb2 as pb
from chalk.features._encoding.json import (
    FeatureEncodingOptions,
    structs_as_objects_feature_json_converter,
    structure_json_to_primitive,
    unstructure_primitive_to_json,
)
from chalk.features._encoding.missing_value import MissingValueStrategy
from chalk.features._encoding.primitive import TPrimitive
from chalk.features._encoding.protobuf import PROTOBUF_TO_UNIT, UNIT_TO_PROTOBUF
from chalk.features._encoding.pyarrow import (
    is_map_in_dtype_tree,
    pyarrow_to_polars,
    pyarrow_to_primitive,
    rich_to_pyarrow,
)
from chalk.features._encoding.rich import structure_primitive_to_rich, unstructure_rich_to_primitive
from chalk.utils.collections import unwrap_annotated_if_needed, unwrap_optional_and_annotated_if_needed
from chalk.utils.df_utils import pa_array_to_pl_series
from chalk.utils.json import JSON, TJSON, is_pyarrow_json_type, pyarrow_json_type
from chalk.utils.log_with_context import get_logger

_logger = get_logger(__name__)

_TRich = TypeVar("_TRich")
_TRichCo = TypeVar("_TRichCo", covariant=True)
_TRichCon = TypeVar("_TRichCon", contravariant=True)

_TPrim = TypeVar("_TPrim", bound=TPrimitive)
_TPrimCo = TypeVar("_TPrimCo", bound=TPrimitive, covariant=True)
_TPrimCon = TypeVar("_TPrimCon", bound=TPrimitive, contravariant=True)

_DEFAULT_FEATURE_ENCODING_OPTIONS = FeatureEncodingOptions()


def _recursively_unwrap(x: Any, dtype: pa.DataType) -> Any:
    if x is None:
        return None
    if isinstance(dtype, pa.MapType):
        if not isinstance(x, list):
            raise ValueError(f"Expected a list, but got {type(x).__name__}")
        if not x:
            return {}
        if not isinstance(x[0], tuple):
            raise TypeError(f"Expected a list of tuples, but got a list of {type(x[0]).__name__}")
        return {k: _recursively_unwrap(v, dtype.item_type) for k, v in x}
    if isinstance(dtype, (pa.ListType, pa.LargeListType, pa.FixedSizeListType)):
        if not isinstance(x, list):
            raise TypeError(f"Expected a list, but got {type(x).__name__}")
        return [_recursively_unwrap(y, dtype.value_type) for y in x]
    if isinstance(dtype, pa.StructType):
        if not isinstance(x, dict):
            raise TypeError(f"Expected a dict, but got {type(x).__name__}")
        return {k: _recursively_unwrap(v, dtype.field(dtype.get_field_index(k)).type) for (k, v) in x.items()}
    if isinstance(x, np.number):
        return x.item()
    return x


class MissingValueError(TypeError):
    """Raised when a missing value is encountered and the missing value strategy is set to ``error``."""

    pass


class PrimitiveFeatureConverter(Generic[_TPrim]):
    """Feature converter that can only deal with primitive types. This means it can be constructed from a serialized graph,
    but it cannot convert to/from rich types"""

    @overload
    def __init__(
        self,
        name: str,
        is_nullable: bool,
        pyarrow_dtype: pa.DataType,
        primitive_default: ellipsis = ...,
    ) -> None:
        ...

    @overload
    def __init__(
        self,
        name: str,
        is_nullable: bool,
        pyarrow_dtype: pa.DataType,
        primitive_default: _TPrim,
    ) -> None:
        ...

    def __init__(
        self,
        name: str,
        is_nullable: bool,
        pyarrow_dtype: pa.DataType,
        primitive_default: Any = ...,
    ) -> None:
        super().__init__()
        self._name = name

        if is_nullable and primitive_default is ...:
            primitive_default = None

        self._pyarrow_dtype = pyarrow_dtype
        if pa.types.is_struct(pyarrow_dtype) and is_nullable:
            is_nullable = False
        self._is_nullable = is_nullable
        self._primitive_type = pyarrow_to_primitive(self._pyarrow_dtype, name)
        try:
            # Storing as an array so we can reuse the same conversion logic
            self._pyarrow_default: ellipsis | pa.Array | pa.ChunkedArray = (
                ... if primitive_default is ... else self.from_primitive_to_pyarrow([primitive_default])
            )
        except Exception as e:
            raise ValueError(
                f"Unable to convert default {primitive_default} of type {type(primitive_default)} to a pyarrow scalar of type {self._pyarrow_dtype}"
            ) from e
        try:
            import polars as pl
        except ModuleNotFoundError:
            pass
        else:
            del pl
            self.polars_dtype: Final = pyarrow_to_polars(self.pyarrow_dtype, self._name)

    __slots__ = ("_name", "_pyarrow_dtype", "_is_nullable", "_primitive_type", "polars_dtype", "_pyarrow_default")

    def __eq__(self, other: object):
        if not isinstance(other, PrimitiveFeatureConverter):
            return NotImplemented
        if not (
            self._name == other._name
            and self._pyarrow_dtype == other._pyarrow_dtype
            and self._is_nullable == other.is_nullable
        ):
            return False
        if self._pyarrow_default == other._pyarrow_default:
            # Fast path -- defaults are definitively equal if 1x1 pyarrow arrays of the default value are equal
            return True
        else:
            if self._pyarrow_default is ... or other._pyarrow_default is ...:
                # Already handled the case where they are both ellipsis above, since ... == ...
                # If here, then one side has a default, when the other side doesn't
                return False
            # Could have false negatives due to nan/null equality
            # To deal, we'll convert both to polars, and compare. This should (recursively) handle nans/nulls
            self_pl = pa_array_to_pl_series(self._pyarrow_default)
            other_pl = pa_array_to_pl_series(other._pyarrow_default)
            import polars.testing

            try:
                # Must use the testing API because otherwise nan != nan
                polars.testing.assert_series_equal(self_pl, other_pl, check_exact=True)
                return True
            except AssertionError:
                return False

    def __hash__(self) -> int:
        return hash((self._name, self._pyarrow_dtype, self._is_nullable, self._pyarrow_default))

    def __getstate__(self):
        # We do NOT include the polars dtype or primitive type on the state; we will reconstruct these from the pyarrow dtype
        # Primitive types might differ if we dynamically construct a class (in the case of structs)
        return (self._name, self._pyarrow_dtype, self._is_nullable, self._pyarrow_default)

    def __setstate__(self, state: tuple[Any, ...]):
        self._name, self._pyarrow_dtype, self._is_nullable, self._pyarrow_default = state
        self._primitive_type = pyarrow_to_primitive(self._pyarrow_dtype, self._name)

        try:
            import polars as pl
        except ModuleNotFoundError:
            pass
        else:
            del pl  # unused
            self.polars_dtype = pyarrow_to_polars(  # pyright: ignore[reportAttributeAccessIssue] -- ok to modify final variables in getstate / setstate
                self.pyarrow_dtype, self._name
            )

    def from_pyarrow_to_json(
        self,
        values: Union[pa.Array, pa.ChunkedArray],
        options: FeatureEncodingOptions = _DEFAULT_FEATURE_ENCODING_OPTIONS,
    ) -> Sequence[TJSON]:
        return [self.from_primitive_to_json(x, options=options) for x in self.from_pyarrow_to_primitive(values)]

    def from_pyarrow_to_primitive(self, values: Union[pa.Array, pa.ChunkedArray]) -> Sequence[_TPrim]:
        return [_recursively_unwrap(x, self._pyarrow_dtype) for x in values.to_pylist()]

    def from_primitive_to_pyarrow(self, value: Iterable[_TPrim]) -> Union[pa.Array, pa.ChunkedArray]:
        # FIXME: Convert recursively
        if pa.types.is_float16(self._pyarrow_dtype):
            return pa.array(np.array([None if x is ... else x for x in value], np.dtype("float16")))
        if pa.types.is_fixed_size_list(self._pyarrow_dtype):
            assert isinstance(self._pyarrow_dtype, pa.FixedSizeListType)
            list_size: int = self._pyarrow_dtype.list_size
            value_type: pa.DataType = self._pyarrow_dtype.value_type
            if pa.types.is_float16(value_type):
                value = tuple(value)
                empty = [None] * list_size
                # NOTE: Numpy converts both nan and null to nan for individual array elements.
                # Because pc.if_else doesn't work for float16 arrays, it is difficult to flip the null bits from the python api
                # Since the main use case of FixedSizeList(float16) is for vectors, will leave the nans as-is, because it is more likely
                # for an individual element of a vector to be nan (i.e. overflow from underlying embedding model) than null
                ans = pa.FixedSizeListArray.from_arrays(
                    np.array([empty if x is ... or x is None else x for x in value], np.dtype("float16")).reshape(-1),
                    list_size,
                )
                mask = [x is not ... and x is not None for x in value]
                ans = pc.if_else(pa.array(mask, pa.bool_()), ans, pa.scalar(None, self._pyarrow_dtype))  # type: ignore
                return ans
        if is_map_in_dtype_tree(self._pyarrow_dtype):
            return pa.array(
                [None if x is ... else self._recursive_dict_to_list_of_dicts(x, self._pyarrow_dtype) for x in value],
                type=self._pyarrow_dtype,
            )
        return pa.array([None if x is ... else x for x in value], type=self._pyarrow_dtype)

    def from_primitive_to_json(
        self, value: TPrimitive, options: FeatureEncodingOptions = _DEFAULT_FEATURE_ENCODING_OPTIONS
    ) -> TJSON:
        if options.encode_structs_as_objects or is_map_in_dtype_tree(self._pyarrow_dtype):
            return structs_as_objects_feature_json_converter.unstructure_primitive_to_json(value)
        else:
            return unstructure_primitive_to_json(value)

    def from_json_to_pyarrow(self, values: Sequence[TJSON]) -> Union[pa.Array, pa.ChunkedArray]:
        primitive_vals = [self.from_json_to_primitive(x) for x in values]
        return self.from_primitive_to_pyarrow(primitive_vals)

    def from_json_to_primitive(self, value: Union[TJSON, TPrimitive]) -> _TPrim:
        try:
            if self._primitive_type == JSON:  # pyright: ignore[reportUnnecessaryComparison]
                # The JSON is the primitive!
                return cast(_TPrim, value)
            return cast(_TPrim, structure_json_to_primitive(value, self._primitive_type))
        except (ValueError, TypeError) as e:
            raise TypeError(f"Could not convert '{value}' to `{self._primitive_type}`: {e}") from e

    @property
    def pyarrow_dtype(self) -> pa.DataType:
        return self._pyarrow_dtype

    @property
    def pyarrow_default(self) -> ellipsis | pa.Array | pa.ChunkedArray:
        return self._pyarrow_default

    @property
    def is_nullable(self) -> bool:
        return self._is_nullable

    @property
    def primitive_type(self) -> Type[TPrimitive]:
        return self._primitive_type

    @property
    def has_default(self) -> bool:
        # primitive default is set <==> rich default is also set or no rich type was provided
        return self._pyarrow_default != ...

    @property
    def primitive_default(self) -> _TPrim:
        if self._pyarrow_default is ...:
            raise ValueError(f"Feature '{self._name}' has no default value")
        return cast(_TPrim, self._pyarrow_default.to_pylist()[0])

    def is_value_missing(self, value: Any):
        """Returns whether the ``value`` should be treated as a "missing" value"""
        if value is ...:
            # Ellipsis is always missing
            return True
        if value is None:
            # All nullable args have a default (``None`` if not otherwise specified)
            if pa.types.is_struct(self.pyarrow_dtype):
                # Nones are not missing for structs
                return False
            return not self._is_nullable
        return False

    @classmethod
    def from_protobuf_to_pyarrow(cls, pb_value: pb.ScalarValue) -> pa.Scalar:
        if pb_value.HasField("extension_value"):
            storage_val = cls.from_protobuf_to_pyarrow(pb_value.extension_value.storage_value)
            logical_type = cls.convert_proto_dtype_to_pa_dtype(
                pb.ArrowType(extension=pb_value.extension_value.extension_type)
            )
            return pa.ExtensionScalar.from_storage(logical_type, storage_val)
        if pb_value.HasField("null_value"):
            return pa.nulls(1, type=cls.convert_proto_dtype_to_pa_dtype(pb_value.null_value))[0]
        if pb_value.HasField("bool_value"):
            return pa.scalar(pb_value.bool_value, pa.bool_())
        if pb_value.HasField("int8_value"):
            return pa.scalar(pb_value.int8_value, pa.int8())
        if pb_value.HasField("int16_value"):
            return pa.scalar(pb_value.int16_value, pa.int16())
        if pb_value.HasField("int32_value"):
            return pa.scalar(pb_value.int32_value, pa.int32())
        if pb_value.HasField("int64_value"):
            return pa.scalar(pb_value.int64_value, pa.int64())
        if pb_value.HasField("uint8_value"):
            return pa.scalar(pb_value.uint8_value, pa.uint8())
        if pb_value.HasField("uint16_value"):
            return pa.scalar(pb_value.uint16_value, pa.uint16())
        if pb_value.HasField("uint32_value"):
            return pa.scalar(pb_value.uint32_value, pa.uint32())
        if pb_value.HasField("uint64_value"):
            return pa.scalar(pb_value.uint64_value, pa.uint64())
        if pb_value.HasField("float16_value"):
            return pa.scalar(np.float16(pb_value.float16_value), pa.float16())
        if pb_value.HasField("float32_value"):
            return pa.scalar(pb_value.float32_value, pa.float32())
        if pb_value.HasField("float64_value"):
            return pa.scalar(pb_value.float64_value, pa.float64())
        if pb_value.HasField("utf8_value"):
            return pa.scalar(pb_value.utf8_value, pa.utf8())
        if pb_value.HasField("large_utf8_value"):
            return pa.scalar(pb_value.large_utf8_value, pa.large_utf8())
        if pb_value.HasField("binary_value"):
            return pa.scalar(pb_value.binary_value, pa.binary())
        if pb_value.HasField("large_binary_value"):
            return pa.scalar(pb_value.large_binary_value, pa.large_binary())
        if pb_value.HasField("fixed_size_binary_value"):
            return pa.scalar(
                pb_value.fixed_size_binary_value.values, pa.binary(pb_value.fixed_size_binary_value.length)
            )
        if pb_value.HasField("date_32_value"):
            return pa.scalar(date(1970, 1, 1) + timedelta(days=pb_value.date_32_value), pa.date32())
        if pb_value.HasField("date_64_value"):
            return pa.scalar(date(1970, 1, 1) + timedelta(days=pb_value.date_64_value), pa.date64())
        if pb_value.HasField("time32_value"):
            if pb_value.time32_value.HasField("time32_second_value"):
                seconds = pb_value.time32_value.time32_second_value
                return pa.scalar(
                    time(
                        hour=seconds // 3600,
                        minute=(seconds % 3600) // 60,
                        second=seconds % 60,
                    ),
                    pa.time32("s"),
                )
            if pb_value.time32_value.HasField("time32_millisecond_value"):
                milliseconds = pb_value.time32_value.time32_millisecond_value
                return pa.scalar(
                    time(
                        hour=milliseconds // 3600000,
                        minute=(milliseconds % 3600000) // 60000,
                        second=(milliseconds % 60000) // 1000,
                        microsecond=(milliseconds % 1000) * 1000,
                    ),
                    pa.time32("ms"),
                )
            raise ValueError(
                "Unsupported time32 value - missing fields `time32_second_value` and `time32_millisecond_value`"
            )
        if pb_value.HasField("time64_value"):
            if pb_value.time64_value.HasField("time64_microsecond_value"):
                microseconds = pb_value.time64_value.time64_microsecond_value
                return pa.scalar(
                    time(
                        hour=microseconds // 3600000000,
                        minute=(microseconds % 3600000000) // 60000000,
                        second=(microseconds % 60000000) // 1000000,
                        microsecond=microseconds % 1000000,
                    ),
                    pa.time64("us"),
                )
            if pb_value.time64_value.HasField("time64_nanosecond_value"):
                nanoseconds = pb_value.time64_value.time64_nanosecond_value
                # Python's datetime.time does not support nanoseconds, so we round to the nearest microsecond
                microseconds = nanoseconds // 1000
                return pa.scalar(
                    time(
                        hour=microseconds // 3600000000,
                        minute=(microseconds % 3600000000) // 60000000,
                        second=(microseconds % 60000000) // 1000000,
                        microsecond=microseconds % 1000000,
                    ),
                    pa.time64("ns"),
                )
            raise ValueError(
                "Unsupported time64 value - missing fields `time64_microsecond_value` and `time64_nanosecond_value`"
            )
        if pb_value.HasField("timestamp_value"):
            tz_str = pb_value.timestamp_value.timezone
            tz = dateutil.tz.gettz(tz_str) if tz_str else None
            if pb_value.timestamp_value.HasField("time_second_value"):
                seconds = pb_value.timestamp_value.time_second_value
                return pa.scalar(datetime.fromtimestamp(seconds, tz=tz), pa.timestamp("s", tz=tz_str))
            if pb_value.timestamp_value.HasField("time_millisecond_value"):
                milliseconds = pb_value.timestamp_value.time_millisecond_value
                return pa.scalar(datetime.fromtimestamp(milliseconds / 1000, tz=tz), pa.timestamp("ms", tz=tz_str))
            if pb_value.timestamp_value.HasField("time_microsecond_value"):
                microseconds = pb_value.timestamp_value.time_microsecond_value
                return pa.scalar(datetime.fromtimestamp(microseconds / 1_000_000, tz=tz), pa.timestamp("us", tz=tz_str))
            if pb_value.timestamp_value.HasField("time_nanosecond_value"):
                nanoseconds = pb_value.timestamp_value.time_nanosecond_value
                return pa.scalar(
                    datetime.fromtimestamp(nanoseconds / 1_000_000_000, tz=tz), pa.timestamp("ns", tz=tz_str)
                )
            raise ValueError(
                (
                    "Unsupported protobuf timestamp value - missing fields `time_second_value`, "
                    "`time_millisecond_value`, `time_microsecond_value`, and `time_nanosecond_value`"
                )
            )
        if pb_value.HasField("struct_value"):
            name_to_pa_scalar = {
                field.name: cls.from_protobuf_to_pyarrow(field_value)
                for field, field_value in zip(pb_value.struct_value.fields, pb_value.struct_value.field_values)
            }
            # TODO: Add test using `v.as_py()` values that evaluate to 0 for testing `if (o := v.as_py()) is not None`
            name_to_py_values_not_none = {k: o for k, v in name_to_pa_scalar.items() if (o := v.as_py()) is not None}
            fields = [pa.field(k, v.type) for k, v in name_to_pa_scalar.items()]
            return pa.scalar(name_to_py_values_not_none, pa.struct(fields))
        if (
            pb_value.HasField("list_value")
            or pb_value.HasField("large_list_value")
            or pb_value.HasField("fixed_size_list_value")
            or pb_value.HasField("map_value")
        ):
            return cls._deserialize_pb_list_to_pa(pb_value)
        if pb_value.HasField("duration_second_value"):
            return pa.scalar(timedelta(seconds=pb_value.duration_second_value), pa.duration("s"))
        if pb_value.HasField("duration_millisecond_value"):
            return pa.scalar(timedelta(milliseconds=pb_value.duration_millisecond_value), pa.duration("ms"))
        if pb_value.HasField("duration_microsecond_value"):
            return pa.scalar(timedelta(microseconds=pb_value.duration_microsecond_value), pa.duration("us"))
        if pb_value.HasField("duration_nanosecond_value"):
            return pa.scalar(timedelta(microseconds=pb_value.duration_nanosecond_value / 1000), pa.duration("ns"))
        if pb_value.HasField("decimal128_value"):
            return cls._deserialize_pb_to_pa_decimal(pb_value)
        if pb_value.HasField("decimal256_value"):
            return cls._deserialize_pb_to_pa_decimal(pb_value)
        raise ValueError(f"Unsupported Protobuf type: {pb_value}")

    @classmethod
    def from_pyarrow_to_protobuf(cls, value: pa.Scalar) -> pb.ScalarValue:
        if value.as_py() is None:
            return pb.ScalarValue(null_value=cls.convert_pa_dtype_to_proto_dtype(value.type))
        if pa.types.is_null(value.type):
            return pb.ScalarValue(null_value=cls.convert_pa_dtype_to_proto_dtype(value.type))
        if pa.types.is_boolean(value.type):
            return pb.ScalarValue(bool_value=value.as_py())
        if pa.types.is_int8(value.type):
            return pb.ScalarValue(int8_value=value.as_py())
        if pa.types.is_int16(value.type):
            return pb.ScalarValue(int16_value=value.as_py())
        if pa.types.is_int32(value.type):
            return pb.ScalarValue(int32_value=value.as_py())
        if pa.types.is_int64(value.type):
            return pb.ScalarValue(int64_value=value.as_py())
        if pa.types.is_uint8(value.type):
            return pb.ScalarValue(uint8_value=value.as_py())
        if pa.types.is_uint16(value.type):
            return pb.ScalarValue(uint16_value=value.as_py())
        if pa.types.is_uint32(value.type):
            return pb.ScalarValue(uint32_value=value.as_py())
        if pa.types.is_uint64(value.type):
            return pb.ScalarValue(uint64_value=value.as_py())
        if pa.types.is_float16(value.type):
            return pb.ScalarValue(float16_value=value.as_py())
        if pa.types.is_float32(value.type):
            return pb.ScalarValue(float32_value=value.as_py())
        if pa.types.is_float64(value.type):
            return pb.ScalarValue(float64_value=value.as_py())
        if pa.types.is_string(value.type):
            return pb.ScalarValue(utf8_value=value.as_py())
        if pa.types.is_large_string(value.type):
            return pb.ScalarValue(large_utf8_value=value.as_py())
        if pa.types.is_binary(value.type):
            return pb.ScalarValue(binary_value=value.as_py())
        if pa.types.is_large_binary(value.type):
            return pb.ScalarValue(large_binary_value=value.as_py())
        if pa.types.is_date32(value.type):
            date_value = value.as_py()
            if not isinstance(date_value, date):
                raise TypeError(f"Expected Python `date` but got `{type(date_value).__name__}`")
            epoch_days = date_value - date(1970, 1, 1)
            return pb.ScalarValue(date_32_value=epoch_days.days)
        if pa.types.is_date64(value.type):
            date_value = value.as_py()
            if not isinstance(date_value, date):
                raise TypeError(f"Expected Python `date` but got `{type(date_value).__name__}`")
            epoch_days = date_value - date(1970, 1, 1)
            return pb.ScalarValue(date_64_value=epoch_days.days)
        if pa.types.is_time32(value.type):
            time_val = value.as_py()
            if not isinstance(time_val, time):
                raise TypeError(f"Expected Python `time`, but got `{type(time_val).__name__}`")
            ms_since_midnight = (
                time_val.hour * 3_600_000
                + time_val.minute * 60_000
                + time_val.second * 1000
                + time_val.microsecond // 1000
            )

            # Failing assertion, because dtype is a `DataType` instead of a `Time32Type`
            # assert isinstance(dtype, pa.Time32Type)
            if value.type == pa.time32("s"):
                return pb.ScalarValue(time32_value=pb.ScalarTime32Value(time32_second_value=ms_since_midnight // 1000))
            assert value.type == pa.time32("ms")
            return pb.ScalarValue(time32_value=pb.ScalarTime32Value(time32_millisecond_value=ms_since_midnight))
        if pa.types.is_time64(value.type):
            time_val = value.as_py()
            if not isinstance(time_val, time):
                raise TypeError(f"Expected Python `time`, but got `{type(time_val).__name__}`")
            ns_since_midnight = (
                time_val.hour * 3_600_000_000_000
                + time_val.minute * 60_000_000_000
                + time_val.second * 1_000_000_000
                + time_val.microsecond * 1000
            )

            # Failing assertion, because dtype is a `DataType` instead of a `Time64Type`
            # assert isinstance(dtype, pa.Time64Type)
            if value.type == pa.time64("us"):
                return pb.ScalarValue(
                    time64_value=pb.ScalarTime64Value(time64_microsecond_value=ns_since_midnight // 1000)
                )
            assert value.type == pa.time64("ns")
            return pb.ScalarValue(time64_value=pb.ScalarTime64Value(time64_nanosecond_value=ns_since_midnight))
        if isinstance(value.type, pa.TimestampType):
            dt_val = value.as_py()
            float_s = dt_val.timestamp()
            if not isinstance(dt_val, datetime):
                raise TypeError(f"Expected Python `datetime`, but got `{type(dt_val).__name__}`")
            timezone = None if dt_val.tzinfo is None else dt_val.tzinfo.tzname(dt_val)

            if value.type.unit == "ms":
                return pb.ScalarValue(
                    timestamp_value=pb.ScalarTimestampValue(
                        time_millisecond_value=int(float_s * 1000),
                        timezone=timezone,
                    )
                )
            elif value.type.unit == "us":
                return pb.ScalarValue(
                    timestamp_value=pb.ScalarTimestampValue(
                        time_microsecond_value=int(float_s * 1_000_000),
                        timezone=timezone,
                    )
                )
            elif value.type.unit == "ns":
                return pb.ScalarValue(
                    timestamp_value=pb.ScalarTimestampValue(
                        time_nanosecond_value=int(float_s * 1_000_000_000),
                        timezone=timezone,
                    )
                )
            return pb.ScalarValue(
                timestamp_value=pb.ScalarTimestampValue(
                    time_second_value=int(float_s),
                    timezone=timezone,
                )
            )
        if pa.types.is_duration(value.type):
            duration_val = value.as_py()
            if not isinstance(duration_val, timedelta):
                raise TypeError(
                    f"Expected a `timedelta` as the Python equivalent of a PyArrow Duration, but got: {type(duration_val).__name__}"
                )
            dtype = value.type
            if not isinstance(dtype, pa.DurationType):
                raise TypeError(
                    f"Expected a `pa.DurationType` as the PyArrow type of a PyArrow Duration, but got: {type(dtype).__name__}"
                )
            if dtype.unit == "s":
                return pb.ScalarValue(duration_second_value=int(duration_val.total_seconds()))
            if dtype.unit == "ms":
                return pb.ScalarValue(duration_millisecond_value=int(duration_val.total_seconds() * 1000))
            if dtype.unit == "us":
                return pb.ScalarValue(duration_microsecond_value=int(duration_val.total_seconds() * 1_000_000))
            if dtype.unit == "ns":
                return pb.ScalarValue(duration_nanosecond_value=int(duration_val.total_seconds() * 1_000_000_000))
            raise ValueError(f"Unsupported duration unit: {dtype.unit}")
        if pa.types.is_fixed_size_binary(value.type):
            bytes_obj = value.as_py()
            if not isinstance(bytes_obj, bytes):
                raise TypeError(f"Expected Python `bytes` but got `{type(bytes_obj).__name__}`")
            return pb.ScalarValue(
                fixed_size_binary_value=pb.ScalarFixedSizeBinary(values=bytes_obj, length=len(bytes_obj))
            )
        if isinstance(value, pa.StructScalar):
            fields = []
            field_values = []
            for name, pa_scalar in value.items():
                pb_scalar = cls.from_pyarrow_to_protobuf(pa_scalar)
                fields.append(
                    pb.Field(
                        name=name,
                        arrow_type=cls.convert_pa_dtype_to_proto_dtype(pa_scalar.type),
                        nullable=True,
                    )
                )
                field_values.append(pb_scalar)
            return pb.ScalarValue(struct_value=pb.StructValue(fields=fields, field_values=field_values))
        if isinstance(value, pa.MapScalar):
            return cls._serialize_pa_list_to_pb(value)
        if isinstance(value, pa.ListScalar):
            return cls._serialize_pa_list_to_pb(value)
        if isinstance(value, pa.Decimal128Scalar):
            return cls._serialize_pa_decimal_to_pb(value)
        if isinstance(value, pa.Decimal256Scalar):
            return cls._serialize_pa_decimal_to_pb(value)

        raise ValueError(f"Unsupported type: {value.type}")

    @classmethod
    def convert_proto_dtype_to_pa_dtype(cls, dtype: pb.ArrowType) -> pa.DataType:
        if dtype.HasField("none"):
            return pa.null()
        elif dtype.HasField("bool"):
            return pa.bool_()
        elif dtype.HasField("int8"):
            return pa.int8()
        elif dtype.HasField("int16"):
            return pa.int16()
        elif dtype.HasField("int32"):
            return pa.int32()
        elif dtype.HasField("int64"):
            return pa.int64()
        elif dtype.HasField("uint8"):
            return pa.uint8()
        elif dtype.HasField("uint16"):
            return pa.uint16()
        elif dtype.HasField("uint32"):
            return pa.uint32()
        elif dtype.HasField("uint64"):
            return pa.uint64()
        elif dtype.HasField("float16"):
            return pa.float16()
        elif dtype.HasField("float32"):
            return pa.float32()
        elif dtype.HasField("float64"):
            return pa.float64()
        elif dtype.HasField("utf8"):
            return pa.utf8()
        elif dtype.HasField("large_utf8"):
            return pa.large_utf8()
        elif dtype.HasField("binary"):
            return pa.binary()
        elif dtype.HasField("large_binary"):
            return pa.large_binary()
        elif dtype.HasField("date32"):
            return pa.date32()
        elif dtype.HasField("date64"):
            return pa.date64()
        elif dtype.HasField("time32"):
            unit = PROTOBUF_TO_UNIT[dtype.time32]
            assert unit == "s" or unit == "ms"
            return pa.time32(unit)
        elif dtype.HasField("time64"):
            unit = PROTOBUF_TO_UNIT[dtype.time64]
            assert unit == "us" or unit == "ns"
            return pa.time64(unit)
        elif dtype.HasField("timestamp"):
            unit = PROTOBUF_TO_UNIT[dtype.timestamp.time_unit]
            assert unit in ("s", "ms", "us", "ns")
            return pa.timestamp(unit, tz=dtype.timestamp.timezone)
        elif dtype.HasField("duration"):
            unit = PROTOBUF_TO_UNIT[dtype.duration]
            assert unit in ("s", "ms", "us", "ns")
            return pa.duration(unit)
        elif dtype.HasField("decimal_128"):
            return pa.decimal128(dtype.decimal_128.precision, dtype.decimal_128.scale)
        elif dtype.HasField("decimal_256"):
            return pa.decimal256(dtype.decimal_256.precision, dtype.decimal_256.scale)
        elif dtype.HasField("struct"):
            fields = [
                pa.field(
                    field.name,
                    cls.convert_proto_dtype_to_pa_dtype(field.arrow_type),
                    nullable=field.nullable,
                )
                for field in dtype.struct.sub_field_types
            ]
            return pa.struct(fields)
        elif dtype.HasField("list"):
            return pa.list_(
                pa.field(
                    name=dtype.list.field_type.name,
                    type=cls.convert_proto_dtype_to_pa_dtype(dtype.list.field_type.arrow_type),
                    nullable=dtype.list.field_type.nullable,
                )
            )
        elif dtype.HasField("large_list"):
            return pa.large_list(
                pa.field(
                    name=dtype.large_list.field_type.name,
                    type=cls.convert_proto_dtype_to_pa_dtype(dtype.large_list.field_type.arrow_type),
                    nullable=dtype.large_list.field_type.nullable,
                )
            )
        elif dtype.HasField("fixed_size_list"):
            return pa.list_(
                pa.field(
                    name=dtype.fixed_size_list.field_type.name,
                    type=cls.convert_proto_dtype_to_pa_dtype(dtype.fixed_size_list.field_type.arrow_type),
                    nullable=dtype.fixed_size_list.field_type.nullable,
                ),
                list_size=dtype.fixed_size_list.list_size,
            )
        elif dtype.HasField("map"):
            key_field = pa.field(
                name=dtype.map.key_field.name,
                type=cls.convert_proto_dtype_to_pa_dtype(dtype.map.key_field.arrow_type),
                nullable=dtype.map.key_field.nullable,
            )
            item_field = pa.field(
                name=dtype.map.item_field.name,
                type=cls.convert_proto_dtype_to_pa_dtype(dtype.map.item_field.arrow_type),
                nullable=dtype.map.item_field.nullable,
            )
            return pa.map_(
                key_type=key_field,
                item_type=item_field,
                keys_sorted=dtype.map.keys_sorted,
            )
        if dtype.HasField("extension"):
            if dtype.extension.name == "arrow.json":
                return pyarrow_json_type()
            raise ValueError(f"Unsupported extension type: '{dtype.extension.name}'")

        raise ValueError(f"Could not convert the protobuf type `{dtype}` to a pyarrow dtype")

    @classmethod
    def convert_pa_dtype_to_proto_dtype(cls, dtype: pa.DataType) -> pb.ArrowType:
        if dtype == pa.null():
            return pb.ArrowType(none=pb.EmptyMessage())
        elif pa.types.is_boolean(dtype):
            return pb.ArrowType(bool=pb.EmptyMessage())
        elif pa.types.is_int8(dtype):
            return pb.ArrowType(int8=pb.EmptyMessage())
        elif pa.types.is_int16(dtype):
            return pb.ArrowType(int16=pb.EmptyMessage())
        elif pa.types.is_int32(dtype):
            return pb.ArrowType(int32=pb.EmptyMessage())
        elif pa.types.is_int64(dtype):
            return pb.ArrowType(int64=pb.EmptyMessage())
        elif pa.types.is_uint8(dtype):
            return pb.ArrowType(uint8=pb.EmptyMessage())
        elif pa.types.is_uint16(dtype):
            return pb.ArrowType(uint16=pb.EmptyMessage())
        elif pa.types.is_uint32(dtype):
            return pb.ArrowType(uint32=pb.EmptyMessage())
        elif pa.types.is_uint64(dtype):
            return pb.ArrowType(uint64=pb.EmptyMessage())
        elif pa.types.is_float16(dtype):
            return pb.ArrowType(float16=pb.EmptyMessage())
        elif pa.types.is_float32(dtype):
            return pb.ArrowType(float32=pb.EmptyMessage())
        elif pa.types.is_float64(dtype):
            return pb.ArrowType(float64=pb.EmptyMessage())
        elif pa.types.is_string(dtype):
            return pb.ArrowType(utf8=pb.EmptyMessage())
        elif pa.types.is_large_string(dtype):
            return pb.ArrowType(large_utf8=pb.EmptyMessage())
        elif pa.types.is_binary(dtype):
            return pb.ArrowType(binary=pb.EmptyMessage())
        elif pa.types.is_large_binary(dtype):
            return pb.ArrowType(large_binary=pb.EmptyMessage())
        elif pa.types.is_date32(dtype):
            return pb.ArrowType(date32=pb.EmptyMessage())
        elif pa.types.is_date64(dtype):
            return pb.ArrowType(date64=pb.EmptyMessage())
        elif dtype == pa.time32("s"):
            return pb.ArrowType(time32=pb.TIME_UNIT_SECOND)
        elif dtype == pa.time32("ms"):
            return pb.ArrowType(time32=pb.TIME_UNIT_MILLISECOND)
        elif dtype == pa.time64("us"):
            return pb.ArrowType(time64=pb.TIME_UNIT_MICROSECOND)
        elif dtype == pa.time64("ns"):
            return pb.ArrowType(time64=pb.TIME_UNIT_NANOSECOND)
        elif isinstance(dtype, pa.TimestampType):
            return pb.ArrowType(timestamp=pb.Timestamp(time_unit=UNIT_TO_PROTOBUF[dtype.unit], timezone=dtype.tz))
        elif isinstance(dtype, pa.DurationType):
            return pb.ArrowType(duration=UNIT_TO_PROTOBUF[dtype.unit])
        elif isinstance(dtype, pa.Decimal128Type):
            return pb.ArrowType(decimal_128=pb.Decimal(precision=dtype.precision, scale=dtype.scale))
        elif isinstance(dtype, pa.Decimal256Type):
            return pb.ArrowType(decimal_256=pb.Decimal(precision=dtype.precision, scale=dtype.scale))
        elif isinstance(dtype, pa.StructType):
            pb_fields = []
            for pa_field in dtype:
                pb_fields.append(
                    pb.Field(
                        name=pa_field.name,
                        nullable=pa_field.nullable,
                        arrow_type=cls.convert_pa_dtype_to_proto_dtype(pa_field.type),
                    )
                )
            return pb.ArrowType(struct=pb.Struct(sub_field_types=pb_fields))
        elif isinstance(dtype, pa.ListType):
            pb_field = pb.Field(
                name=dtype.value_field.name,
                nullable=dtype.value_field.nullable,
                arrow_type=cls.convert_pa_dtype_to_proto_dtype(dtype.value_field.type),
            )
            return pb.ArrowType(list=pb.List(field_type=pb_field))
        elif isinstance(dtype, pa.LargeListType):
            pb_field = pb.Field(
                name=dtype.value_field.name,
                nullable=dtype.value_field.nullable,
                arrow_type=cls.convert_pa_dtype_to_proto_dtype(dtype.value_field.type),
            )
            return pb.ArrowType(large_list=pb.List(field_type=pb_field))
        elif isinstance(dtype, pa.FixedSizeListType):
            pb_field = pb.Field(
                name=dtype.value_field.name,
                nullable=dtype.value_field.nullable,
                arrow_type=cls.convert_pa_dtype_to_proto_dtype(dtype.value_field.type),
            )
            return pb.ArrowType(fixed_size_list=pb.FixedSizeList(field_type=pb_field, list_size=dtype.list_size))
        elif isinstance(dtype, pa.MapType):
            return pb.ArrowType(
                map=pb.Map(
                    key_field=pb.Field(
                        name=dtype.key_field.name,
                        nullable=dtype.key_field.nullable,
                        arrow_type=cls.convert_pa_dtype_to_proto_dtype(dtype.key_field.type),
                    ),
                    item_field=pb.Field(
                        name=dtype.item_field.name,
                        nullable=dtype.item_field.nullable,
                        arrow_type=cls.convert_pa_dtype_to_proto_dtype(dtype.item_field.type),
                    ),
                ),
            )
        elif isinstance(dtype, pa.ExtensionType):
            return pb.ArrowType(
                extension=pb.Extension(
                    name=dtype.extension_name,
                    storage_type=cls.convert_pa_dtype_to_proto_dtype(dtype.storage_type),
                )
            )
        else:
            raise TypeError(f"Could not convert the pyarrow dtype {dtype} to a protobuf message")

    @staticmethod
    def _serialize_pa_decimal_to_pb(value: Union[pa.Decimal128Scalar, pa.Decimal256Scalar]) -> pb.ScalarValue:
        dec_val = value.as_py()
        if not isinstance(dec_val, Decimal):
            raise ValueError(
                f"Expected a `Decimal` as the Python equivalent of a PyArrow decimal, but got a '{type(dec_val).__name__}'"
            )

        dtype = value.type
        if not isinstance(dtype, (pa.Decimal128Type, pa.Decimal256Type)):
            raise TypeError(
                f"Expected a `pa.Decimal128Type` or `pa.Decimal256Type` as the PyArrow type of a PyArrow decimal, but got: {type(dtype).__name__}"
            )

        _, digits, exponent = dec_val.as_tuple()
        scale = int(exponent * -1)
        if scale != dtype.scale:
            raise ValueError(
                f"Expected Python Decimal {dec_val} to have a scale of {dtype.scale}, but got scale {scale}"
            )
        int_val = int(dec_val.scaleb(Decimal(scale)))
        bytes_val = int_val.to_bytes((int_val.bit_length() + 7) // 8, byteorder="big", signed=True)
        precision = max(len(digits), scale)
        if precision != dtype.precision:
            raise ValueError(
                f"Expected Python Decimal {dec_val} to have a precision of {dtype.precision}, but got precision {precision}"
            )

        if isinstance(value, pa.Decimal256Scalar):
            return pb.ScalarValue(decimal256_value=pb.DecimalValue(value=bytes_val, precision=precision, scale=scale))

        return pb.ScalarValue(decimal128_value=pb.DecimalValue(value=bytes_val, precision=precision, scale=scale))

    @staticmethod
    def _deserialize_pb_to_pa_decimal(scalar_value: pb.ScalarValue) -> pa.Scalar:
        if scalar_value.HasField("decimal128_value"):
            value = scalar_value.decimal128_value
        elif scalar_value.HasField("decimal256_value"):
            value = scalar_value.decimal256_value
        else:
            raise ValueError("Protobuf ScalarValue does not contain a Decimal128 or a Decimal256 value")
        bytes_val = value.value
        expected_precision = value.precision
        scale = value.scale

        # Convert bytes back to integer
        int_val = int.from_bytes(bytes_val, byteorder="big", signed=True)

        # Construct Decimal from integer
        decimal_scale = Decimal(10) ** -scale
        dec_val = Decimal(int_val) * decimal_scale

        # Validate precision
        _, digits, exponent = dec_val.as_tuple()
        assert isinstance(exponent, int)
        actual_precision = max(len(digits), exponent * -1)
        if actual_precision != expected_precision:
            raise ValueError(f"Reconstructed Decimal has precision {expected_precision}; expected {actual_precision}")

        if scalar_value.HasField("decimal256_value"):
            dtype = pa.decimal256(actual_precision, scale)
        else:
            dtype = pa.decimal128(actual_precision, scale)
        return pa.scalar(dec_val, type=dtype)

    @classmethod
    def _serialize_pa_list_to_pb(cls, value: pa.ListScalar) -> pb.ScalarValue:
        values = value.values
        if values is None:
            return pb.ScalarValue(null_value=cls.convert_pa_dtype_to_proto_dtype(value.type))
        table = pa.Table.from_arrays([values], names=["values"])

        arrow_buffer = BytesIO()
        pf.write_feather(table, dest=arrow_buffer, compression=None)
        arrow_bytes = arrow_buffer.getvalue()

        pb_schema = pb.Schema(
            columns=[
                pb.Field(
                    nullable=False,
                    arrow_type=cls.convert_pa_dtype_to_proto_dtype(values.type),
                ),
            ]
        )

        pb_value = pb.ScalarListValue(
            arrow_data=arrow_bytes,
            schema=pb_schema,
        )
        if isinstance(value, pa.LargeListScalar):
            return pb.ScalarValue(large_list_value=pb_value)
        elif isinstance(value, pa.FixedSizeListScalar):
            return pb.ScalarValue(fixed_size_list_value=pb_value)
        elif isinstance(value, pa.MapScalar):
            return pb.ScalarValue(map_value=pb_value)
        elif isinstance(value, pa.ListScalar):
            return pb.ScalarValue(list_value=pb_value)
        else:
            raise TypeError(f"Unsupported list type: {type(value).__name__}")

    @classmethod
    def _deserialize_pb_list_to_pa(cls, scalar_value: pb.ScalarValue) -> pa.Scalar:
        if scalar_value.HasField("list_value"):
            value = scalar_value.list_value
        elif scalar_value.HasField("large_list_value"):
            value = scalar_value.large_list_value
        elif scalar_value.HasField("fixed_size_list_value"):
            value = scalar_value.fixed_size_list_value
        elif scalar_value.HasField("map_value"):
            value = scalar_value.map_value
        else:
            raise ValueError(
                "Protobuf ScalarValue does not contain the attributes "
                + "`list_value`, `large_list_value`, `map_value`, or `fixed_size_list_value`"
            )

        if not isinstance(value, pb.ScalarListValue):  # pyright: ignore[reportUnnecessaryIsInstance]
            raise TypeError(f"Unsupported list value type: {type(value).__name__}")

        if value.arrow_data:
            arrow_buffer = BytesIO(value.arrow_data)
            table = pf.read_table(arrow_buffer)
        else:
            raise ValueError("Missing `arrow_data` attribute in `ScalarListValue`")

        arr = table.column(0).combine_chunks()

        if scalar_value.HasField("list_value"):
            return pa.scalar(arr.to_pylist(), pa.list_(arr.type))
        elif scalar_value.HasField("large_list_value"):
            return pa.scalar(arr.to_pylist(), pa.large_list(arr.type))
        elif scalar_value.HasField("map_value"):
            if not isinstance(arr.type, pa.StructType):
                raise TypeError(f"Expected a `Struct` but got: {type(arr).__name__}")
            return pa.scalar(
                arr.to_pylist(),
                pa.map_(key_type=arr.type.field("key"), item_type=arr.type.field("value")),
            )
        else:
            return pa.scalar(arr.to_pylist(), pa.list_(arr.type, len(arr)))

    @classmethod
    def _recursive_dict_to_list_of_dicts(cls, d: Any, dtype: pa.DataType) -> Any:
        if d is None:
            return None
        if isinstance(dtype, pa.MapType):
            if isinstance(d, pa.MapScalar):
                # If the value is already wrapped as an Arrow scalar, unwrap it.
                d = {k: v for k, v in d.as_py()}

            if not isinstance(d, dict):
                raise TypeError(f"Expected a `dict` but got: {type(d).__name__}")
            return [{"key": k, "value": cls._recursive_dict_to_list_of_dicts(v, dtype.item_type)} for k, v in d.items()]
        elif isinstance(dtype, (pa.ListType, pa.LargeListType, pa.FixedSizeListType)):
            return [cls._recursive_dict_to_list_of_dicts(x, dtype.value_type) for x in d]
        elif isinstance(dtype, pa.StructType):
            if not isinstance(d, dict):
                raise TypeError(f"Expected a `dict` but got: {type(d).__name__}")
            res = {}
            for k, v in d.items():
                field_idx = dtype.get_field_index(k)
                if field_idx == -1:
                    raise ValueError(f"Struct type does not have a field named '{k}'")
                res[k] = cls._recursive_dict_to_list_of_dicts(v, dtype.field(field_idx).type)
            return res
        return d

    def from_primitive_to_protobuf(
        self,
        value: _TPrim,
    ) -> pb.ScalarValue:
        if isinstance(value, pa.Scalar):
            # This value is alreay a scalar.
            as_scalar = value
        else:
            as_arr = self.from_primitive_to_pyarrow([value])
            as_scalar = as_arr[0]
        return self.from_pyarrow_to_protobuf(as_scalar)


class TEncoder(Protocol[_TPrimCo, _TRichCon]):
    def __call__(self, value: _TRichCon, /) -> _TPrimCo:
        ...


class TDecoder(Protocol[_TPrimCon, _TRichCo]):
    def __call__(self, value: _TPrimCon, /) -> _TRichCo:
        ...


def _encode_json(t: JSON) -> str:
    return json.dumps(t)


def _decode_json(s: str) -> JSON:
    return cast(JSON, json.loads(s))


def _to_old_style_type(origin: object):
    if origin is list:
        return List
    elif origin is tuple:
        return Tuple
    elif origin is set:
        return Set
    elif origin is frozenset:
        return FrozenSet
    elif origin is dict:
        return Dict
    elif hasattr(types, "UnionType"):
        # Only available in python >=3.10
        if origin is getattr(types, "UnionType"):
            return Union
    return origin


def _canonicalize_typ(x: object):
    """Canonicalize a type annotation for equality checking.
    New-style types are replaced with old-style types, and
    annotated markers are ignored. Specifically:

    - typing.Annotated -> unwrapped
    - list             -> typing.List
    - tuple            -> typing.Tuple
    - set              -> typing.Set
    - frozenset        -> typing.Frozenset
    - dict             -> typing.Dict
    - union            -> typing.Union
    """

    if x is None:
        x = type(None)
    x = unwrap_annotated_if_needed(x)
    origin, args = get_origin(x), get_args(x)
    if origin is None:
        return _to_old_style_type(x)
    origin = _to_old_style_type(origin)
    args = tuple(_canonicalize_typ(x) for x in args)
    return origin[args]  # type: ignore -- pyright doesn't understand metaprogramming


@final
class JSONCodec:
    encode: TEncoder[str, JSON] = _encode_json
    decode: TDecoder[str, JSON] = _decode_json


class FeatureConverter(PrimitiveFeatureConverter[_TPrim], Generic[_TPrim, _TRich]):
    """Feature converter that deals with rich types. It supports everything that the primitive feature converter supports.

    However, since it deals with Rich types, it can only be constructed from the source code, since rich type information
    is not stored in serialized graphs."""

    def __init__(
        self,
        name: str,
        is_nullable: bool,
        rich_type: Union[Type[_TRich], ellipsis] = ...,
        primitive_default: Union[_TPrim, ellipsis] = ...,
        rich_default: Union[_TRich, ellipsis] = ...,
        pyarrow_dtype: Optional[pa.DataType] = None,
        encoder: Optional[TEncoder[_TPrim, _TRich]] = None,
        decoder: Optional[TDecoder[_TPrim, _TRich]] = None,
    ) -> None:
        self._rich_type = unwrap_annotated_if_needed(rich_type)

        if pyarrow_dtype is None:
            if rich_type is ...:
                raise ValueError("Either the `rich_type` or `pyarrow_dtype` must be provided")
            pyarrow_dtype = rich_to_pyarrow(rich_type, name)

        if rich_type is ...:
            if rich_default != ...:
                raise ValueError(
                    "The `rich_default` cannot be used without the `rich_type`. Perhaps specify the `primitive_default` instead?"
                )
            if is_nullable and primitive_default is ...:
                primitive_default = cast(_TPrim, None)

        else:
            if primitive_default != ...:
                raise ValueError(
                    "The `primitive_default` cannot be used when specifying the `rich_type`. Instead, specify the `rich_default`."
                )
            if is_nullable and rich_default is ...:
                rich_default = cast(_TRich, None)

        # In the future, we will require the rich type to be not-none and remove the primitive default flag,
        # and then we can simplify the code as follows:
        # if pyarrow_dtype is None:
        #     if rich_type is ...:
        #         raise ValueError("Either the `rich_type` or `pyarrow_dtype` must be provided")
        #     pyarrow_dtype = rich_to_pyarrow(rich_type, name)

        # elif is_nullable and rich_default is ...:
        #     rich_default = cast(_TRich, None)

        if rich_type is ...:
            if encoder is not None:
                raise ValueError("An encoder cannot be specified without also specifying the `rich_type`")
            if decoder is not None:
                raise ValueError("An encoder cannot be specified without also specifying the `rich_type`")
        self._encoder = encoder
        self._decoder = decoder
        self._rich_default = rich_default
        self._primitive_type = pyarrow_to_primitive(pyarrow_dtype, name)
        self._pyarrow_dtype = pyarrow_dtype
        self._is_nullable = is_nullable

        # This field is also set in the super() call, but must be initialized here
        # because it is also used for error handling inside of `from_rich_to_primitive`.
        self._name = name
        if rich_default != ...:
            # The missing value strategy doesn't really matter because rich_default is not missing
            primitive_default = self.from_rich_to_primitive(rich_default, missing_value_strategy="allow")
        super().__init__(
            name, is_nullable=is_nullable, pyarrow_dtype=pyarrow_dtype, primitive_default=primitive_default
        )

    @property
    def rich_type(self) -> Type[_TRich]:
        if self._rich_type is ...:
            raise ValueError(
                "Rich types cannot be used as the FeatureConverter was created without providing a `rich_type`"
            )
        return cast(Type[_TRich], self._rich_type)

    @property
    def rich_default(self) -> _TRich:
        if self._rich_default is ...:
            raise ValueError(f"Feature '{self._name}' has no default value")
        return self._rich_default

    def is_rich_valid(self, value: _TRich) -> bool:
        """Returns true if value has a valid rich type"""
        try:
            prim = self.from_rich_to_primitive(value, "default_or_error")
            pa.scalar(prim, type=self.pyarrow_dtype)
            return True
        except (TypeError, ValueError):
            return False

    def from_rich_to_pyarrow(
        self,
        values: Sequence[Union[_TRich, ellipsis, None]],
        /,
        missing_value_strategy: MissingValueStrategy = "default_or_allow",
    ) -> Union[pa.Array, pa.ChunkedArray]:
        prim_values = [self.from_rich_to_primitive(x, missing_value_strategy) for x in values]
        return self.from_primitive_to_pyarrow(prim_values)

    def from_rich_to_protobuf(
        self,
        value: Union[_TRich, ellipsis, None],
        missing_value_strategy: MissingValueStrategy = "default_or_allow",
    ) -> pb.ScalarValue:
        return self.from_primitive_to_protobuf(self.from_rich_to_primitive(value, missing_value_strategy))

    def from_rich_to_primitive(
        self,
        value: Union[_TRich, ellipsis, None],
        missing_value_strategy: MissingValueStrategy = "default_or_allow",
    ) -> _TPrim:
        # Ensure that the rich value is indeed the rich type
        # For example, if a string is passed in for a datetime value, convert it into a datetime
        if self.is_value_missing(value):
            if missing_value_strategy == "allow":
                return cast(_TPrim, value)
            elif missing_value_strategy in ("default_or_error", "default_or_allow"):
                if self.has_default:
                    return self.primitive_default
                elif missing_value_strategy == "default_or_error":
                    raise TypeError(
                        f"The value for feature '{self._name}' is missing, and this feature has no default value."
                    )
                else:
                    return cast(_TPrim, value)
            elif missing_value_strategy == "error":
                raise MissingValueError(
                    f"The value for feature '{self._name}' is missing, but `replace_missing_with_defaults` was set to False."
                )
            else:
                raise ValueError(
                    (
                        f"Unsupported missing value strategy: {missing_value_strategy}. "
                        "It must be one of 'allow', 'default_or_allow', 'default_or_error', or 'error'."
                    )
                )
        value = self.from_primitive_to_rich(cast(_TPrim, value))
        return self._to_primitive(value)

    def from_rich_to_json(
        self,
        value: Union[_TRich, ellipsis, None],
        missing_value_strategy: MissingValueStrategy = "default_or_allow",
        options: FeatureEncodingOptions = _DEFAULT_FEATURE_ENCODING_OPTIONS,
    ) -> TJSON:
        prim_val = self.from_rich_to_primitive(value, missing_value_strategy)
        return self.from_primitive_to_json(prim_val, options=options)

    def from_pyarrow_to_rich(self, values: Union[pa.Array, pa.ChunkedArray], /) -> Sequence[_TRich]:
        return [self.from_primitive_to_rich(x) for x in self.from_pyarrow_to_primitive(values)]

    @property
    def encoder(self) -> Optional[TEncoder[_TPrim, _TRich]]:
        return self._encoder

    @property
    def decoder(self) -> Optional[TDecoder[_TPrim, _TRich]]:
        return self._decoder

    def _to_primitive(self, val: _TRich) -> _TPrim:
        if val is None or self._encoder is None:
            # Structuring null values to the primitive type to ensure that a singular null for an entire struct
            # is propagated to individual struct fields -- e.g.
            # class LatLong:
            #     lat: Optional[float]
            #     long: Optional[float]
            # then self._from_prim(None) == LatLong(None, None)
            # Using self.primitive_type, rather than self._rich_type, as the primitive type
            # might not be registered on the converter for custom classes
            try:
                x = unstructure_rich_to_primitive(val)
            except (TypeError, ValueError) as e:
                raise TypeError(
                    f"Could not convert '{val}' to `{self.primitive_type}` for feature '{self._name}': {e}"
                ) from e
            if x is None and not self._is_nullable:
                raise ValueError(f"Feature '{self._name}' is null, but it cannot be nullable")
            try:
                return cast(_TPrim, structure_primitive_to_rich(x, self.primitive_type))
            except (TypeError, ValueError) as e:
                raise TypeError(
                    f"Could not convert '{val}' to `{self.primitive_type}` for feature '{self._name}': {e}"
                ) from e
        return self._encoder(val)

    def _from_prim(self, val: Union[_TPrim, _TRich]) -> _TRich:
        if self._rich_type is ...:
            raise ValueError(
                "Rich types cannot be used as the FeatureConverter was created without providing a `rich_type`"
            )
        if val is None:
            # Structuring null values to the primitive type to ensure that a singular null for an entire struct
            # is propagated to individual struct fields -- e.g.
            # class LatLong:
            #     lat: Optional[float]
            #     long: Optional[float]
            # then self._from_prim(None) == LatLong(None, None)
            # Using self.primitive_type, rather than self._rich_type, as the primitive type
            # might not be registered on the converter for custom classes
            try:
                val = structure_primitive_to_rich(cast(_TPrim, val), cast(Type[_TRich], self.primitive_type))
            except (TypeError, ValueError) as e:
                raise TypeError(
                    f"Could not convert '{val}' to `{self.primitive_type}` for feature '{self._name}': {e}"
                ) from e
        if self._decoder is None:
            try:
                return structure_primitive_to_rich(cast(_TPrim, val), self._rich_type)
            except (TypeError, ValueError) as e:
                raise TypeError(
                    f"Could not convert '{val}' to `{self._rich_type}` for feature '{self._name}': {e}"
                ) from e
        # is_pyarrow_json_type is only needed to handle python 3.8
        if is_pyarrow_json_type(self.pyarrow_dtype) or isinstance(
            val, unwrap_optional_and_annotated_if_needed(self._rich_type)
        ):
            return cast(_TRich, val)
        if val is None:
            # If the value is None, then we won't call the custom converter, since those likely cannot handle null values
            # and None is perfectly valid as a "rich" type
            return cast(_TRich, None)
        return self._decoder(cast(_TPrim, val))

    def from_primitive_to_rich(self, value: Union[_TPrim, _TRich]) -> _TRich:
        return self._from_prim(value)

    def from_json_to_rich(self, value: TJSON) -> _TRich:
        prim_val = self.from_json_to_primitive(value)
        return self.from_primitive_to_rich(prim_val)

    def has_nontrivial_rich_type(self) -> bool:
        if self._encoder is not None or self._decoder is not None:
            return True
        return _canonicalize_typ(self.primitive_type) != _canonicalize_typ(self.rich_type)



import os

_TRUTHY_VALUES = {"1", "true", "yes", "t", "y"}
_FALSY_VALUES = {"0", "false", "no", "f", "n"}


def env_var_bool(env_name: str, default: bool = False) -> bool:
    var = os.environ.get(env_name)
    if var is None:
        return default
    if var.lower() in _TRUTHY_VALUES:
        return True
    if var.lower() in _FALSY_VALUES:
        return False
    raise ValueError(
        f"Environ '{env_name}' is set to '{var}', which cannot be converted to a boolean value. Please set to '0' or '1'."
    )



from chalk._gen.chalk.auth.v1 import permissions_pb2 as _permissions_pb2
from google.protobuf.internal import containers as _containers
from google.protobuf.internal import enum_type_wrapper as _enum_type_wrapper
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from typing import (
    ClassVar as _ClassVar,
    Iterable as _Iterable,
    Mapping as _Mapping,
    Optional as _Optional,
    Union as _Union,
)

DESCRIPTOR: _descriptor.FileDescriptor

class FlagScope(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
    __slots__ = ()
    FLAG_SCOPE_UNSPECIFIED: _ClassVar[FlagScope]
    FLAG_SCOPE_TEAM: _ClassVar[FlagScope]
    FLAG_SCOPE_ENVIRONMENT: _ClassVar[FlagScope]

FLAG_SCOPE_UNSPECIFIED: FlagScope
FLAG_SCOPE_TEAM: FlagScope
FLAG_SCOPE_ENVIRONMENT: FlagScope

class FeatureFlagValue(_message.Message):
    __slots__ = ("flag", "value")
    FLAG_FIELD_NUMBER: _ClassVar[int]
    VALUE_FIELD_NUMBER: _ClassVar[int]
    flag: str
    value: bool
    def __init__(self, flag: _Optional[str] = ..., value: bool = ...) -> None: ...

class GetFeatureFlagsRequest(_message.Message):
    __slots__ = ()
    def __init__(self) -> None: ...

class GetFeatureFlagsResponse(_message.Message):
    __slots__ = ("flags",)
    FLAGS_FIELD_NUMBER: _ClassVar[int]
    flags: _containers.RepeatedCompositeFieldContainer[FeatureFlagValue]
    def __init__(self, flags: _Optional[_Iterable[_Union[FeatureFlagValue, _Mapping]]] = ...) -> None: ...

class SetFeatureFlagRequest(_message.Message):
    __slots__ = ("flag", "value", "scope")
    FLAG_FIELD_NUMBER: _ClassVar[int]
    VALUE_FIELD_NUMBER: _ClassVar[int]
    SCOPE_FIELD_NUMBER: _ClassVar[int]
    flag: str
    value: bool
    scope: FlagScope
    def __init__(
        self, flag: _Optional[str] = ..., value: bool = ..., scope: _Optional[_Union[FlagScope, str]] = ...
    ) -> None: ...

class SetFeatureFlagResponse(_message.Message):
    __slots__ = ()
    def __init__(self) -> None: ...



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: chalk/server/v1/datasets.proto
# Protobuf Python Version: 4.25.3
"""Generated protocol buffer code."""

from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from chalk._gen.chalk.auth.v1 import permissions_pb2 as chalk_dot_auth_dot_v1_dot_permissions__pb2
from google.protobuf import struct_pb2 as google_dot_protobuf_dot_struct__pb2
from google.protobuf import timestamp_pb2 as google_dot_protobuf_dot_timestamp__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(
    b'\n\x1e\x63halk/server/v1/datasets.proto\x12\x0f\x63halk.server.v1\x1a\x1f\x63halk/auth/v1/permissions.proto\x1a\x1cgoogle/protobuf/struct.proto\x1a\x1fgoogle/protobuf/timestamp.proto"\xb6\n\n\x13\x44\x61tasetRevisionMeta\x12\x1d\n\nnumeric_id\x18\x01 \x01(\x03R\tnumericId\x12(\n\x10offline_query_id\x18\x02 \x01(\tR\x0eofflineQueryId\x12\x1d\n\ndataset_id\x18\x03 \x01(\tR\tdatasetId\x12"\n\ngivens_uri\x18\x04 \x01(\tH\x00R\tgivensUri\x88\x01\x01\x12V\n\x0egivens_version\x18\x05 \x01(\x0e\x32*.chalk.server.v1.OfflineQueryGivensVersionH\x01R\rgivensVersion\x88\x01\x01\x12"\n\noutput_uri\x18\x06 \x01(\tH\x02R\toutputUri\x88\x01\x01\x12K\n\x0eoutput_version\x18\x07 \x01(\x0e\x32\x1f.chalk.server.v1.DatasetVersionH\x03R\routputVersion\x88\x01\x01\x12$\n\x0b\x62ranch_name\x18\x08 \x01(\tH\x04R\nbranchName\x88\x01\x01\x12\x1e\n\x08num_rows\x18\t \x01(\x03H\x05R\x07numRows\x88\x01\x01\x12\x33\n\x13physical_size_bytes\x18\n \x01(\x03H\x06R\x11physicalSizeBytes\x88\x01\x01\x12%\n\x0eoutput_columns\x18\x0b \x03(\tR\routputColumns\x12\x1f\n\x0boutput_fqns\x18\x0c \x03(\tR\noutputFqns\x12\x1e\n\x08\x61gent_id\x18\r \x01(\tH\x07R\x07\x61gentId\x88\x01\x01\x12\x42\n\x0c\x63ompleted_at\x18\x0e \x01(\x0b\x32\x1a.google.protobuf.TimestampH\x08R\x0b\x63ompletedAt\x88\x01\x01\x12"\n\nnum_shards\x18\x0f \x01(\x03H\tR\tnumShards\x88\x01\x01\x12(\n\rnum_computers\x18\x10 \x01(\x03H\nR\x0cnumComputers\x88\x01\x01\x12\x37\n\x08metadata\x18\x11 \x01(\x0b\x32\x16.google.protobuf.ValueH\x0bR\x08metadata\x88\x01\x01\x12\x43\n\x06status\x18\x12 \x01(\x0e\x32&.chalk.server.v1.DatasetRevisionStatusH\x0cR\x06status\x88\x01\x01\x12\x33\n\x13num_rows_calculated\x18\x13 \x01(\x03H\rR\x11numRowsCalculated\x88\x01\x01\x12H\n\x1ephysical_size_bytes_calculated\x18\x14 \x01(\x03H\x0eR\x1bphysicalSizeBytesCalculated\x88\x01\x01\x12>\n\ncreated_at\x18\x15 \x01(\x0b\x32\x1a.google.protobuf.TimestampH\x0fR\tcreatedAt\x88\x01\x01\x42\r\n\x0b_givens_uriB\x11\n\x0f_givens_versionB\r\n\x0b_output_uriB\x11\n\x0f_output_versionB\x0e\n\x0c_branch_nameB\x0b\n\t_num_rowsB\x16\n\x14_physical_size_bytesB\x0b\n\t_agent_idB\x0f\n\r_completed_atB\r\n\x0b_num_shardsB\x10\n\x0e_num_computersB\x0b\n\t_metadataB\t\n\x07_statusB\x16\n\x14_num_rows_calculatedB!\n\x1f_physical_size_bytes_calculatedB\r\n\x0b_created_at"\xae\x02\n\x0b\x44\x61tasetMeta\x12\x0e\n\x02id\x18\x01 \x01(\tR\x02id\x12%\n\x0e\x65nvironment_id\x18\x02 \x01(\tR\renvironmentId\x12&\n\x0c\x64\x61taset_name\x18\x03 \x01(\tH\x00R\x0b\x64\x61tasetName\x88\x01\x01\x12\x39\n\ncreated_at\x18\x05 \x01(\x0b\x32\x1a.google.protobuf.TimestampR\tcreatedAt\x12[\n\x14most_recent_revision\x18\x07 \x01(\x0b\x32$.chalk.server.v1.DatasetRevisionMetaH\x01R\x12mostRecentRevision\x88\x01\x01\x42\x0f\n\r_dataset_nameB\x17\n\x15_most_recent_revision"\x8a\x01\n\x13ListDatasetsRequest\x12\x1b\n\x06\x63ursor\x18\x01 \x01(\tH\x00R\x06\x63ursor\x88\x01\x01\x12\x19\n\x05limit\x18\x02 \x01(\x05H\x01R\x05limit\x88\x01\x01\x12\x1b\n\x06search\x18\x03 \x01(\tH\x02R\x06search\x88\x01\x01\x42\t\n\x07_cursorB\x08\n\x06_limitB\t\n\x07_search"x\n\x14ListDatasetsResponse\x12\x38\n\x08\x64\x61tasets\x18\x01 \x03(\x0b\x32\x1c.chalk.server.v1.DatasetMetaR\x08\x64\x61tasets\x12\x1b\n\x06\x63ursor\x18\x02 \x01(\tH\x00R\x06\x63ursor\x88\x01\x01\x42\t\n\x07_cursor"#\n\x11GetDatasetRequest\x12\x0e\n\x02id\x18\x01 \x01(\tR\x02id"L\n\x12GetDatasetResponse\x12\x36\n\x07\x64\x61taset\x18\x01 \x01(\x0b\x32\x1c.chalk.server.v1.DatasetMetaR\x07\x64\x61taset"\x89\x01\n\x1bListDatasetRevisionsRequest\x12\x1d\n\ndataset_id\x18\x01 \x01(\tR\tdatasetId\x12\x1b\n\x06\x63ursor\x18\x02 \x01(\tH\x00R\x06\x63ursor\x88\x01\x01\x12\x19\n\x05limit\x18\x03 \x01(\x05H\x01R\x05limit\x88\x01\x01\x42\t\n\x07_cursorB\x08\n\x06_limit"\x8a\x01\n\x1cListDatasetRevisionsResponse\x12\x42\n\trevisions\x18\x01 \x03(\x0b\x32$.chalk.server.v1.DatasetRevisionMetaR\trevisions\x12\x1b\n\x06\x63ursor\x18\x02 \x01(\tH\x00R\x06\x63ursor\x88\x01\x01\x42\t\n\x07_cursor"+\n\x19GetDatasetRevisionRequest\x12\x0e\n\x02id\x18\x01 \x01(\tR\x02id"^\n\x1aGetDatasetRevisionResponse\x12@\n\x08revision\x18\x01 \x01(\x0b\x32$.chalk.server.v1.DatasetRevisionMetaR\x08revision*\x9f\x02\n\x15\x44\x61tasetRevisionStatus\x12\'\n#DATASET_REVISION_STATUS_UNSPECIFIED\x10\x00\x12#\n\x1f\x44\x41TASET_REVISION_STATUS_UNKNOWN\x10\x01\x12#\n\x1f\x44\x41TASET_REVISION_STATUS_WORKING\x10\x02\x12%\n!DATASET_REVISION_STATUS_COMPLETED\x10\x03\x12"\n\x1e\x44\x41TASET_REVISION_STATUS_FAILED\x10\x04\x12$\n DATASET_REVISION_STATUS_CANCELED\x10\x05\x12"\n\x1e\x44\x41TASET_REVISION_STATUS_QUEUED\x10\x06*\xe9\x02\n\x0e\x44\x61tasetVersion\x12\x1f\n\x1b\x44\x41TASET_VERSION_UNSPECIFIED\x10\x00\x12\x1b\n\x17\x44\x41TASET_VERSION_UNKNOWN\x10\x01\x12:\n6DATASET_VERSION_BIGQUERY_JOB_WITH_B32_ENCODED_COLNAMES\x10\x02\x12"\n\x1e\x44\x41TASET_VERSION_DATASET_WRITER\x10\x03\x12=\n9DATASET_VERSION_BIGQUERY_JOB_WITH_B32_ENCODED_COLNAMES_V2\x10\x04\x12.\n*DATASET_VERSION_COMPUTE_RESOLVER_OUTPUT_V1\x10\x05\x12!\n\x1d\x44\x41TASET_VERSION_NATIVE_DTYPES\x10\x06\x12\'\n#DATASET_VERSION_NATIVE_COLUMN_NAMES\x10\x07*\xad\x02\n\x19OfflineQueryGivensVersion\x12,\n(OFFLINE_QUERY_GIVENS_VERSION_UNSPECIFIED\x10\x00\x12(\n$OFFLINE_QUERY_GIVENS_VERSION_UNKNOWN\x10\x01\x12>\n:OFFLINE_QUERY_GIVENS_VERSION_NATIVE_TS_FEATURE_FOR_ROOT_NS\x10\x02\x12\x33\n/OFFLINE_QUERY_GIVENS_VERSION_SINGLE_TS_COL_NAME\x10\x03\x12\x43\n?OFFLINE_QUERY_GIVENS_VERSION_SINGLE_TS_COL_NAME_WITH_URI_PREFIX\x10\x04\x32\xd0\x03\n\x16\x44\x61tasetMetadataService\x12\x63\n\x0cListDatasets\x12$.chalk.server.v1.ListDatasetsRequest\x1a%.chalk.server.v1.ListDatasetsResponse"\x06\x90\x02\x01\x80}\x04\x12]\n\nGetDataset\x12".chalk.server.v1.GetDatasetRequest\x1a#.chalk.server.v1.GetDatasetResponse"\x06\x90\x02\x01\x80}\x04\x12{\n\x14ListDatasetRevisions\x12,.chalk.server.v1.ListDatasetRevisionsRequest\x1a-.chalk.server.v1.ListDatasetRevisionsResponse"\x06\x90\x02\x01\x80}\x04\x12u\n\x12GetDatasetRevision\x12*.chalk.server.v1.GetDatasetRevisionRequest\x1a+.chalk.server.v1.GetDatasetRevisionResponse"\x06\x90\x02\x01\x80}\x04\x42\x96\x01\n\x13\x63om.chalk.server.v1B\rDatasetsProtoP\x01Z\x12server/v1;serverv1\xa2\x02\x03\x43SX\xaa\x02\x0f\x43halk.Server.V1\xca\x02\x0f\x43halk\\Server\\V1\xe2\x02\x1b\x43halk\\Server\\V1\\GPBMetadata\xea\x02\x11\x43halk::Server::V1b\x06proto3'
)

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, "chalk.server.v1.datasets_pb2", _globals)
if _descriptor._USE_C_DESCRIPTORS == False:
    _globals["DESCRIPTOR"]._options = None
    _globals[
        "DESCRIPTOR"
    ]._serialized_options = b"\n\023com.chalk.server.v1B\rDatasetsProtoP\001Z\022server/v1;serverv1\242\002\003CSX\252\002\017Chalk.Server.V1\312\002\017Chalk\\Server\\V1\342\002\033Chalk\\Server\\V1\\GPBMetadata\352\002\021Chalk::Server::V1"
    _globals["_DATASETMETADATASERVICE"].methods_by_name["ListDatasets"]._options = None
    _globals["_DATASETMETADATASERVICE"].methods_by_name["ListDatasets"]._serialized_options = b"\220\002\001\200}\004"
    _globals["_DATASETMETADATASERVICE"].methods_by_name["GetDataset"]._options = None
    _globals["_DATASETMETADATASERVICE"].methods_by_name["GetDataset"]._serialized_options = b"\220\002\001\200}\004"
    _globals["_DATASETMETADATASERVICE"].methods_by_name["ListDatasetRevisions"]._options = None
    _globals["_DATASETMETADATASERVICE"].methods_by_name[
        "ListDatasetRevisions"
    ]._serialized_options = b"\220\002\001\200}\004"
    _globals["_DATASETMETADATASERVICE"].methods_by_name["GetDatasetRevision"]._options = None
    _globals["_DATASETMETADATASERVICE"].methods_by_name[
        "GetDatasetRevision"
    ]._serialized_options = b"\220\002\001\200}\004"
    _globals["_DATASETREVISIONSTATUS"]._serialized_start = 2590
    _globals["_DATASETREVISIONSTATUS"]._serialized_end = 2877
    _globals["_DATASETVERSION"]._serialized_start = 2880
    _globals["_DATASETVERSION"]._serialized_end = 3241
    _globals["_OFFLINEQUERYGIVENSVERSION"]._serialized_start = 3244
    _globals["_OFFLINEQUERYGIVENSVERSION"]._serialized_end = 3545
    _globals["_DATASETREVISIONMETA"]._serialized_start = 148
    _globals["_DATASETREVISIONMETA"]._serialized_end = 1482
    _globals["_DATASETMETA"]._serialized_start = 1485
    _globals["_DATASETMETA"]._serialized_end = 1787
    _globals["_LISTDATASETSREQUEST"]._serialized_start = 1790
    _globals["_LISTDATASETSREQUEST"]._serialized_end = 1928
    _globals["_LISTDATASETSRESPONSE"]._serialized_start = 1930
    _globals["_LISTDATASETSRESPONSE"]._serialized_end = 2050
    _globals["_GETDATASETREQUEST"]._serialized_start = 2052
    _globals["_GETDATASETREQUEST"]._serialized_end = 2087
    _globals["_GETDATASETRESPONSE"]._serialized_start = 2089
    _globals["_GETDATASETRESPONSE"]._serialized_end = 2165
    _globals["_LISTDATASETREVISIONSREQUEST"]._serialized_start = 2168
    _globals["_LISTDATASETREVISIONSREQUEST"]._serialized_end = 2305
    _globals["_LISTDATASETREVISIONSRESPONSE"]._serialized_start = 2308
    _globals["_LISTDATASETREVISIONSRESPONSE"]._serialized_end = 2446
    _globals["_GETDATASETREVISIONREQUEST"]._serialized_start = 2448
    _globals["_GETDATASETREVISIONREQUEST"]._serialized_end = 2491
    _globals["_GETDATASETREVISIONRESPONSE"]._serialized_start = 2493
    _globals["_GETDATASETREVISIONRESPONSE"]._serialized_end = 2587
    _globals["_DATASETMETADATASERVICE"]._serialized_start = 3548
    _globals["_DATASETMETADATASERVICE"]._serialized_end = 4012
# @@protoc_insertion_point(module_scope)



from chalk._gen.chalk.lsp.v1 import lsp_pb2 as pb
from chalk.parsed.duplicate_input_gql import (
    CodeActionGQL,
    CodeDescriptionGQL,
    DiagnosticGQL,
    DiagnosticRelatedInformationGQL,
    DiagnosticSeverityGQL,
    LocationGQL,
    LspGQL,
    PositionGQL,
    PublishDiagnosticsParams,
    RangeGQL,
    TextDocumentEditGQL,
    TextDocumentIdentifierGQL,
    TextEditGQL,
    WorkspaceEditGQL,
)

_severity_to_proto = {
    DiagnosticSeverityGQL.Error: pb.DIAGNOSTIC_SEVERITY_ERROR,
    DiagnosticSeverityGQL.Warning: pb.DIAGNOSTIC_SEVERITY_WARNING,
    DiagnosticSeverityGQL.Information: pb.DIAGNOSTIC_SEVERITY_INFORMATION,
    DiagnosticSeverityGQL.Hint: pb.DIAGNOSTIC_SEVERITY_HINT,
}


def convert_diagnostic_params(params: PublishDiagnosticsParams) -> pb.PublishDiagnosticsParams:
    return pb.PublishDiagnosticsParams(diagnostics=[convert_diagnostic(d) for d in params.diagnostics], uri=params.uri)


def convert_position(position: PositionGQL) -> pb.Position:
    return pb.Position(line=position.line, character=position.character)


def convert_range(range: RangeGQL) -> pb.Range:
    return pb.Range(start=convert_position(range.start), end=convert_position(range.end))


def convert_diagnostic_severity(severity: DiagnosticSeverityGQL) -> pb.DiagnosticSeverity:
    res = _severity_to_proto.get(severity)
    if res is None:
        raise ValueError(f"Unknown diagnostic severity: {severity}")
    return res


def convert_code_description(description: CodeDescriptionGQL) -> pb.CodeDescription:
    return pb.CodeDescription(href=description.href)


def convert_location(location: LocationGQL) -> pb.Location:
    return pb.Location(uri=location.uri, range=convert_range(location.range))


def convert_related_information(info: DiagnosticRelatedInformationGQL) -> pb.DiagnosticRelatedInformation:
    return pb.DiagnosticRelatedInformation(location=convert_location(info.location), message=info.message)


def convert_diagnostic(diagnostic: DiagnosticGQL) -> pb.Diagnostic:
    return pb.Diagnostic(
        range=convert_range(diagnostic.range),
        message=diagnostic.message,
        severity=convert_diagnostic_severity(diagnostic.severity) if diagnostic.severity is not None else None,
        code=diagnostic.code,
        code_description=convert_code_description(diagnostic.codeDescription)
        if diagnostic.codeDescription is not None
        else None,
        related_information=[convert_related_information(r) for r in diagnostic.relatedInformation or []],
    )


def convert_text_document(identifier: TextDocumentIdentifierGQL) -> pb.TextDocumentIdentifier:
    return pb.TextDocumentIdentifier(uri=identifier.uri)


def convert_text_edit(edit: TextEditGQL) -> pb.TextEdit:
    return pb.TextEdit(range=convert_range(edit.range), new_text=edit.newText)


def convert_document_edit(edit: TextDocumentEditGQL) -> pb.TextDocumentEdit:
    return pb.TextDocumentEdit(
        text_document=convert_text_document(edit.textDocument),
        edits=[convert_text_edit(e) for e in edit.edits],
    )


def convert_workspace_edit(edit: WorkspaceEditGQL) -> pb.WorkspaceEdit:
    return pb.WorkspaceEdit(
        document_changes=[convert_document_edit(c) for c in edit.documentChanges],
    )


def convert_code_action(action: CodeActionGQL) -> pb.CodeAction:
    return pb.CodeAction(
        title=action.title,
        diagnostics=[convert_diagnostic(d) for d in action.diagnostics or []],
        edit=convert_workspace_edit(action.edit),
    )


def convert_lsp_gql_to_proto(lsp: LspGQL) -> pb.LSP:
    return pb.LSP(
        diagnostics=[convert_diagnostic_params(p) for p in lsp.diagnostics],
        actions=[convert_code_action(a) for a in lsp.actions],
    )



"""
@generated by mypy-protobuf.  Do not edit manually!
isort:skip_file
"""

from abc import (
    ABCMeta,
    abstractmethod,
)
from chalk._gen.chalk.server.v1.metrics_pb2 import (
    GetOverviewSummaryMetricsRequest,
    GetOverviewSummaryMetricsResponse,
)
from grpc import (
    Channel,
    Server,
    ServicerContext,
    UnaryUnaryMultiCallable,
)

class MetricsServiceStub:
    def __init__(self, channel: Channel) -> None: ...
    GetOverviewSummaryMetrics: UnaryUnaryMultiCallable[
        GetOverviewSummaryMetricsRequest,
        GetOverviewSummaryMetricsResponse,
    ]

class MetricsServiceServicer(metaclass=ABCMeta):
    @abstractmethod
    def GetOverviewSummaryMetrics(
        self,
        request: GetOverviewSummaryMetricsRequest,
        context: ServicerContext,
    ) -> GetOverviewSummaryMetricsResponse: ...

def add_MetricsServiceServicer_to_server(servicer: MetricsServiceServicer, server: Server) -> None: ...
